
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <gmodule.h>

#include "db.h"
#include "ui.h"
#include "bc.h"

#include "bond.h"
#include "bondmain.h"
#include "bondtrigger.h"
#include "watchtower.h"

/**
 * bond_init:
 * @argc: From your main function ... (incomplete) 
 * @argv:
 * @bond:
 * 
 * bond_init() will look for a bond.conf which it will read for
 * configuration information about both the database to connect to and
 * the glade file to read in.  Call this function in your main() file to
 * initialise everything.
 *
 * Returns: non-zero on error
 */

gint
bond_init(gint argc, gchar * argv[], Bond ** bond)
	{
	gint err = 0;
	static gchar defaultplugin[] = "uigtk";
	gchar *bondxmlfile = NULL, *uiplugin;
	gchar *linklibrary, *linkinitfunc;
	gint i = 0;

	if (argc < 1)
		exit( -6);

	if (bc_init(argc, argv))
		{
		error_output("No configuration file. Aborting.\n");
		exit( -1);
		}
	if (db_init(NULL))
		{
		error_output("Failed to connect to database. Aborting.");
		exit( -2);
		}
	uiplugin = bcconfig_getvalue("uiplugin");
	if (uiplugin == NULL)
		uiplugin = defaultplugin;

	/* bondxml file be expressed on the commandline to overrides value in conf
	   file */
	for (i = 1; i < argc; i++)
		{
		if (!strcmp(argv[i], "-x"))
			bondxmlfile = g_strdup(argv[i + 1]);
		}

	if (bondxmlfile == NULL)
		{
		bondxmlfile = bcconfig_getvalue("bondxml");
		if (bondxmlfile == NULL)
			{
			error_output
			("No bond xml document is found, please define one in your config file.\n");
			exit( -3);
			}
		}

	/* initialise bond */
	*bond = mem_calloc(sizeof(Bond));
	(*bond)->uixml = ui_init(uiplugin, NULL, bondxmlfile, argc, argv);
	if (!(*bond)->uixml)
		{
		error_output("Failed to initalise user interface. Aborting\n");
		exit( -4);
		}	
	(*bond)->uixml->parent = (*bond);
	debug_output("User interface is now initialised.\n");
	debug_line(NULL, NULL, Debug_Line);

	/* if successful err should remain 0, else it will increase */
	err -= bond_inithash(*bond);
	if (!err)
		debug_output("Bond repository has been created. Preparing "
			     "to parse the xml document.\n");
	ui_parse((*bond)->uixml);
	debug_output("First parse of %s is successful.\n", bondxmlfile);

	err -= bondxml_init(*bond);
	if (!err)
		debug_output("Second parse of %s is successful.\n", bondxmlfile);

	linklibrary = bcconfig_getvalue("loadmodule");
	if (linklibrary)
		{
		linkinitfunc = bcconfig_getvalue("loadmodule_initfunction");
		if (!bond_linkagainst((*bond), linklibrary, linkinitfunc))
			debug_output("Loading of module %s is successful.\n", linklibrary);
		else
			error_output("Failed to load module %s.\n", linklibrary);
		}

	err -= bondtrigger_init(*bond);
	if (!err)
		debug_output("Callback triggers are set.\n");

	if (err)
		{
		error_output("Failed to initalise bond properly. Aborting\n");
		mem_free(*bond);
		exit( -6);
		}
	registration_init(*bond);
	debug_output("Registry is loaded and ready.\n");
	debug_line(NULL, NULL, Debug_Line);

	return 0;
	}

static void
bond_saveall_element(void *key, void *value, void *ptr)
	{
	GList *walk, *bobjlist;

	bobjlist = value;
	for (walk = g_list_first(value); walk != NULL; walk = walk->next)
		bond_save(walk->data, TRUE, FALSE);
	}

/**
 * bond_saveall:
 * @bond: Global bond object
 * 
 * Save all windows currently open.
 */
void
bond_saveall(Bond * bond)
	{
	g_hash_table_foreach(bond->globalbobj_byobj, bond_saveall_element, NULL);
	}

static void
bond_closeall_element(void *key, void *value, void *ptr)
	{
	GList *toclose;
	BondObject *bobj;

	bobj = value;
	toclose = ptr;
	/* Only returns parent windows to close */
	if (bobj->parent == NULL)
		{
		toclose = g_list_append(toclose, bobj);
		toclose = g_list_append(toclose, bobj->widgetname);
		}
	}

/**
 * bond_closeall:
 * @bond: Global bond object.
 * 
 * Close all windows currently open and save there contents. This is automatically
 * called when you call bond_cleanup().
 */
void
bond_closeall(Bond * bond)
	{
	gchar *widgetname;
	BondObject *bobj;
	GList *walk;
	GList *toclose;

	/* first element is NULL so i dont have to do nasty pointer stuff */
	toclose = g_list_prepend(NULL, NULL);

	/* get a list of all bondobjects to close */
	g_hash_table_foreach(bond->globalbobj, bond_closeall_element, toclose);
	/* first item skipped cause its NULL */
	walk = g_list_first(toclose);
	walk = walk->next;
	for (; walk != NULL; walk = walk->next)
		{
		bobj = walk->data;
		walk = walk->next;
		widgetname = walk->data;
		/* Do a quick check to make certain we dont already have this open */
		if (bond_getbyname(bond, widgetname) == NULL)
			continue;
		/* Save contents, close and free the bond object */
		bond_close(bobj, TRUE);
		}
	g_list_free(toclose);
	}

/**
 * bond_cleanup:
 *
 * Frees up and cleans away everything once your done. call only once. 
 * you can't be too clean can you?
 */
gint
bond_cleanup(Bond * bond)
	{
	GList *walk;

	g_assert(bond);
	debug_line(NULL, NULL, Debug_Line);
	debug_output("Cleanup for exit.\n");
	mem_checkboundaries();

	debug_output("All open forms hare are been saved back to the database.\n");
	bond_closeall(bond);
	debug_output("Flushing database connection.\n");
	db_toliet_flushall();
	for (walk = g_list_first(bond->othermodules); walk != NULL;
	        walk = walk->next)
		g_module_close(walk->data);
	if (bond->othermodules)
		g_list_free(bond->othermodules);
	g_module_close(bond->symbols);
	debug_output("Dynamic modules have now been unloaded.\n");

	bond_cleanuphash(bond);
	debug_output("BondObjects have been freed.\n");
	registration_cleanup(bond);
	debug_output("Registration of bond objects have been canceled.\n");
	bondxml_cleanup(bond);
	debug_output("XML document have now been unloaded.\n");

	ui_cleanup(bond->uixml);
	debug_output("User interface have been disactiviated.\n");
	mem_free(bond);
	db_cleanup();
	debug_output("Database base connections are now closed.\n");

	debug_line(NULL, NULL, Debug_Line);
	fflush(NULL);
	debug_output("GoodBye\n");
	bc_cleanup();
	fflush(NULL);

	return 0;
	}

static void
bond_showall_windows(void *key_widget, void *value_widgetname, void *ptr)
	{
	BondObject *bobj;
	Bond *bond = ptr;

	if (ui_gettype(bond->uixml, key_widget) != UiForm)
		return ;
	if (ui_isvisible(bond->uixml, key_widget) != TRUE)
		return ;

	debuglog(30, "Displaying %s with a bit of guesswork.", value_widgetname);
	bobj = bond_spawn(bond, NULL, value_widgetname, NULL, FALSE);
	bond_paint(bobj, FALSE, TRUE);
	}

/**
 * bond_showall:
 *
 * Does a load all on every visible form currently on the screen
 *
 * Returns: non-zero on error
 */
gint
bond_showall(Bond * bond)
	{
	g_assert(bond);
	debug_output("All visible windows are being painted.\n");
	g_hash_table_foreach(bond->uixml->widgets, bond_showall_windows, bond);

	return 0;
	}

/**
 * bond_main:
 *
 * Enter the main loop of your programme.  Run this after you shown and set up
 * any forms you need to enter the main loop.
 */
gint
bond_main(Bond * bond)
	{
	g_assert(bond);
	debug_line(NULL, NULL, Debug_Line);
	debug_output("Entering main loop.\n");
	return ui_main(bond->uixml);
	}

/**
 * bond_linkagainst:
 * @library: Dynamic library to link against.
 * @initfunc: Initalising function to call in what we link against.
 *
 * This will cause bond to dynamically add another library at runtime
 * and call teh initfunc within it.  This is highly useful if you want
 * to have a library containing your callback functions and you want
 * to link against it at runtime.
 *
 * Returns: non-zero on error.
 */
gint
bond_linkagainst(Bond * bond, gchar * library, gchar * initfunc)
	{
	GModule *mod;
	void *ptr;
	typedef gint(*function) (Bond *);
	function func;

	g_assert(bond);
	g_assert(library);

	if (!g_module_supported())
		{
		errormsg
		("The application needs g_module support before it can have signals");
		return -1;
		}

	mod = (GModule *) g_module_open(library, 0);
	if (mod == NULL)
		{
		error_output
		("Failed to load dynamic module %s. Please ensure the library is present.\n",
		 library);
		return -2;
		}
	/* Call initalisation function if needed */
	if (initfunc)
		{
		if (g_module_symbol(mod, initfunc, &ptr))
			{
			func = ptr;
			func(bond);
			}
		else
			error_output
			("Unable to find function %s specified in the config file.\n",
			 initfunc);
		}

	bond->othermodules = g_list_prepend(bond->othermodules, mod);
	return 0;
	}
