/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <strings.h>
#include <glib.h>

#include "bc.h"
#include "db.h"
#include "dbfilter.h"

#include "bond.h"
#include "bondrelation.h"
#include "bondfilter.h"
#include "search.h"
#include "watchtower.h"
#include "evaluate.h"

/**
 * bondrelation_create:
 * @type: The BondType enum. (normally use RelationAuto)
 * 
 * Returns: Newly created BondRelation. 
 */
BondRelation *
bondrelation_create(RelationType type)
	{
	BondRelation *brel;

	brel = (BondRelation *) mem_calloc(sizeof(BondRelation));
	brel->type = type;
	brel->refcounter = 1;
	return brel;
	}

/**
 * bondrelation_free:
 * @brel: BondRelation to free.
 * 
 * Frees a @brel.
 * 
 * Returns: non-zero on error.
 */
gint
bondrelation_free(BondRelation * brel)
	{
	mem_verify(brel);

	brel->refcounter--;
	if (brel->refcounter == 0)
		{
		if (brel->parentobj)
			g_list_free(brel->parentobj);
		if (brel->loadbywidget)
			g_list_free(brel->loadbywidget);
		mem_free(brel);
		return 0;
		}
	return 1;
	}

/* make dbobj refer to null to all instances of @obj. */
/* This will look up all bond objects and anything that refers to @oldobj will
 * be changed to refer to @newobj. */
static gint
bondrelation_changeobjs(Bond * bond, DbObject * newobj, DbObject * oldobj)
	{
	gint retval = -2;
	GList *objlist, *walk, *newobjlist = NULL;
	BondObject *bobj;

	g_assert(bond);
	/* abort if oldobj is null. ie new spawn of obj. */
	if (oldobj == NULL)
		return 1;
	/* abort if newobj already registered - should never happen */
	if (newobj && g_hash_table_lookup(bond->globalbobj_byobj, newobj))
		{
		errormsg("Oh no an dbobject of same ptr already exists. must "
		         "abort");
		g_assert(NULL);
		return -3;
		}
	/* get all references. */
	objlist = g_hash_table_lookup(bond->globalbobj_byobj, oldobj);

	for (walk = g_list_first(objlist); walk != NULL; walk = walk->next)
		{
		bobj = walk->data;
		bobj->obj = newobj;
		/* If its a new object (not a blank case) then re-register */
		if (newobj)
			newobjlist = g_list_prepend(newobjlist, bobj);
		retval = 0;
		}
	g_hash_table_remove(bond->globalbobj_byobj, oldobj);
	if (newobjlist)
		g_hash_table_insert(bond->globalbobj_byobj, newobj, newobjlist);
	return retval;
	}

static void
bondrelation_freecache(void *ptr)
	{
	db_freeobject((DbObject *) ptr);
	}

static DbObject *
bondrelation_checkcache(BondObject * bobj)
	{
	DbObject *obj = NULL;

	/* check if i am suppose to cache this */
	if (bobj->oprnode->dbobj->cache == FALSE)
		return NULL;
	/* create hash table if not exist */
	if (bobj->bond->objcache == NULL)
		bobj->bond->objcache = g_hash_table_new_full(g_str_hash,
		                       g_str_equal, NULL,
		                       bondrelation_freecache);
	/* find if object is already in cache */
	obj = g_hash_table_lookup(bobj->bond->objcache, bobj->widgetname);
	if (obj == NULL)
		{
		obj = bondrelation_load(bobj->bond, NULL, bobj->brel);
		g_hash_table_insert(bobj->bond->objcache, bobj->widgetname, obj);
		}
	debuglog(80, "I've used cached version of dbobject in %s", bobj->widgetname);
	bobj->obj = obj;
	return obj;
	}

/* Guess how the data is suppose to be loaded for this dbobject. Ie is it a sql
 * statement or a straight table. This converts the RelationAuto to something 
 * meanful */
static RelationType
bondrelation_guess(OperationsNodeDb * dbobj)
	{
	RelationType type = RelationNone;

	/* check for loadall flag */
	if (dbobj->loadall == TRUE)
		return RelationLoadAll;

	/* if (bobj->oprnode->dbobj->loadby != NULL && bobj->oprnode->dbobj->sql !=
	   NULL) return RelationLoadBySQL; */
	if (dbobj->sql != NULL)
		return RelationSQL;

	/* if (bobj->oprnode->dbobj->loadby != NULL) return RelationLoadBy; */
	/* last resort, make a blank object to it */
	if (dbobj->table != NULL)
		return RelationAdd;
	/* oh thats it i give up */
	return type;
	}

static void
bondrelation_sqlchange(void *ptr1, void *ptr2)
	{
	WatchTower *watchtower;
	BondObject *watched, *redraw;
	OperationsNodeFilter *oprfilter;
	Object *objchanged;
	gchar *value = NULL;
	DbFilterAppend *filter;

	/* initialise everything */
	watchtower = ptr1;
	watched = ptr2;
	attention("Calling sql change! ");
	errormsg(" CODE DISABLED!");
#if 0
	/* Convert the dependon BondObjects to a list of dbobject's that these
	   dependon's are derived from. */
	if (brel->bobj->dependon)
		{
		db_clearobject(obj);
		db_setready(obj, TRUE);
		ignoreparentbrel = TRUE;

		for (walk = g_list_first(brel->bobj->dependon); walk != NULL;
		        walk = walk->next)
			{
			tmpbobj = walk->data;
			if (!tmpbobj || !tmpbobj->dbobjectsrc)
				continue;
			tmpobj = tmpbobj->dbobjectsrc->obj;
			message("Adding %s dbobject from %s",
			        tmpbobj->dbobjectsrc->widgetname, tmpbobj->widgetname);
			if (tmpobj)
				brel->parentobj = g_list_append(brel->parentobj, tmpobj);
			else
				warningmsg("Can not find a database object associed with %s",
				           walk->data);
			}
		}
#endif
	}

static gint
bondrelation_buildwatchtower(BondObject * bobjwatching, GList * dependon)
	{
	gint retval = 0;
	GList *walk;
	BondObject *toredraw;
	WatchTower *watchtower;
	EvaluateDependOn *depend;

	mem_verify(bobjwatching);
	/* toredraw is normally same as bobjwatching, but in case of dropdown boxes
	   it could be different because of the lookup/dropdown are different
	   widgets, so you want to redraw the dropdown not the lookup */
	toredraw = bobjwatching;
	if (bobjwatching->widget == NULL)
		if (bobjwatching->parent && bobjwatching->parent->widget)
			toredraw = bobjwatching->parent;

	for (walk = g_list_first(dependon); walk != NULL; walk = walk->next)
		{
		depend = walk->data;
		if (depend->bobj == NULL)
			{
			/* Todo -- Queue to build watchtower */
			}
		watchtower = watchtower_create(depend->bobj, bondrelation_sqlchange);
		watchtower->ptr[0] = toredraw;
		if (depend->type == DependWidget)
			{
			/* If widget doesn't exist at time of creation relook for it
			   just to be sure */
			if (depend->widget == NULL)
				depend->widget =
				    ui_getwidget(bobjwatching->bond->uixml,
				                 depend->widgetname);
			watchtower->widget = depend->widget;
			}
		else if (depend->type == DependField)
			{
			watchtower->tablename = depend->tablename;
			watchtower->fieldname = depend->fieldname;
			watchtower->dbobjectsrc = depend->dbobjectsrc;
			}
		}
	watchtower_register(watchtower);

	return retval;
	}

/**
 * bondrelation_load:
 * @bond: Global bond object.
 * @obj: Orginal db object
 * @brel: A BondRelation to use for loading the data up.
 * 
 * Load the actual records from the database.
 * 
 * Returns: A newly created dbobj or a modified @obj if @obj wasn't %NULL.
 */
DbObject *
bondrelation_load(Bond * bond, DbObject * obj, BondRelation * brel)
	{
	gint err = 0;
	gboolean ignoreparentbrel = FALSE;
	gchar *classname, *tmpstr;
	GList *dependon;
	DbObject *tmpobj;
	BondObject *tmpbobj;
	GList *walk;

	mem_verify(brel);
	mem_verify(brel->dbobj);

	if (brel->dontrunyet == TRUE)
		{
		brel->dontrunyet = FALSE;
		return obj;
		}
	/* for creating basic type. */
	classname = brel->dbobj->table;
	if (obj == NULL)
		obj = db_createobject(classname);

	/* if i am to automatically set the relationship */
	if (brel->type == RelationAuto)
		brel->type = bondrelation_guess(brel->dbobj);
	/* if we have to do a search, mark the object for it */
	if (brel->dbobj->search || brel->dbobj->loadby || brel->parentobj)
		{
		db_setready(obj, FALSE);
		if (brel->type == RelationNone || brel->type == RelationAdd)
			{
			tmpstr = mem_strdup_printf("SELECT * FROM %s", classname);
			db_sql(obj, tmpstr);
			mem_free(tmpstr);
			}
		}
	/* act on the type now */
	switch (brel->type)
		{
		/* No relation with a database object */
	case RelationNone:
		debuglog(60, "Relation is set to RelationNone");
		break;

		/* Add a record to a table. */
	case RelationAdd:
		if (classname == NULL)
			{
			errormsg("No classname found for loading %s",
			         brel->dbobj->widgetname);
			return NULL;
			}
		break;

	case RelationLoadAll:
		debuglog(60, "Loading all records in %s", classname);
		if (classname == NULL)
			{
			errormsg("No classname found for loading %s",
			         brel->dbobj->widgetname);
			return NULL;
			}
		db_loadall(obj);
		if (err < 0)
			warningmsg("db_loadall(); failed with %d", err);
		break;

	case RelationSQL:
		debuglog(60, "Running sql statement on %s",
		         brel->dbobj->widgetname);
		if (brel->sql)
			tmpstr = bond_evaluate(brel->bobj, brel->sql, &dependon);
		else if (brel->dbobj->sql)
			tmpstr = bond_evaluate(brel->bobj, brel->dbobj->sql, &dependon);
		else
			{
			errormsg("No SQL statement specified for loading %s",
			         brel->dbobj->widgetname);
			return NULL;
			}
		/* Convert the dependon BondObjects to a list of dbobject's that
		 * these dependon's are derived from. */
		if (dependon)
			{
			if (!brel->bobj)
				errormsg("No bondobject specified so aboring");
			else
				bondrelation_buildwatchtower(brel->bobj, dependon);

			db_clearobject(obj);
			db_setready(obj, TRUE);
			ignoreparentbrel = TRUE;

			for (walk = g_list_first(brel->bobj->dependon); 
				walk != NULL; walk = walk->next)
				{
				tmpbobj = walk->data;
				if (!tmpbobj || !tmpbobj->dbobjectsrc)
					continue;
				tmpobj = tmpbobj->dbobjectsrc->obj;
				message("Adding %s dbobject from %s",
				        tmpbobj->dbobjectsrc->widgetname,
				        tmpbobj->widgetname);
				if (tmpobj)
					brel->parentobj =
					    g_list_append(brel->parentobj, tmpobj);
				else
					warningmsg("Can not find a database "
					           "object associed with %s", 
						   walk->data);
				}
			}
		/* Save any objects we depend on first */
		/* bondrelation_saveeffected(bond, brel->parentobj); */
		/* Run the sql */
		debuglog(40, "sql is : %s", tmpstr);
		obj = db_sql(obj, tmpstr);
		mem_free(tmpstr);
		/* m.essage("%d results ready %d on
		   %s",db_numrecord(obj),obj->sqlready,brel->bobj->widgetname); */
		break;

	default:
		errormsg("Unknown relationship type");
		break;
		}

	/* Handle any loadby's set in the bond relation. It will only 
	 * load the ones specified in the xml file once, and ignore 
	 * them there after. */
	if ((brel->dbobj->loadby || brel->parentobj) && ignoreparentbrel == FALSE)
		{
		debuglog(60, "Constructiong a loadby join on %s", obj->basetable);
		if (brel->parentobjloaded == FALSE)
			for (walk = g_list_first(brel->dbobj->loadby);
			        walk != NULL; walk = walk->next)
				{
				tmpobj = bondrelation_findobjfromobjsrc(
						bond, walk->data);
				if (tmpobj)
					brel->parentobj = g_list_append(
							brel->parentobj, tmpobj);
				else
					warningmsg("Can not find a database "
					           "object associed with %s", 
						   walk->data);
				}
		brel->parentobjloaded = TRUE;

		if (brel->parentobj == NULL)
			{
			errormsg("No parent dbobject present, aborting loadby.");
			return NULL;
			}
		err -= bondrelation_saveeffected(bond, brel->parentobj);
		if (err < 0)
			warningmsg("bondrelation_saveeffected(); "
				"failed with %d", err);
		/* Load db obj by a list of parent db objects previously 
		 * generated */
		if (db_loadobjectbyobjects(obj, brel->parentobj) == -2)
			{
			if (brel->bobj)
				warningmsg("db_loadobjectbyobjects() failed "
					"with -2 on %s.", err, 
					brel->bobj->widgetname);
			err -= -2;
			/* db_obj_setdead(obj,TRUE); */
			}
		}

	if (err != 0)
		{
		warningmsg("Error occured well loading object. err=%d", err);
		return obj;
		}
	g_assert(obj);
	/* if a search has to be done, do it */
	if (brel->dbobj->search)
		{
		err = search_parse(bond, brel, obj);
		/* If you want to mix a search and a loadall 
		 * then it can't run this code */
		if (err == 1 && brel->type != RelationLoadAll)
			{
			db_obj_clear(obj);
			obj->sqlready = TRUE;
			}
		}

	if (err >= 0)
		if (brel->dbobj->search || brel->dbobj->loadby)
			db_setready(obj, TRUE);
	/* This is normally ok unless it is error -4 */
	if (db_wakemastersoftables(obj) < 0)
		debugmsg("wakemasteroftable failed with %d",
		           db_wakemastersoftables(obj));
	return obj;
	}

/**
 * bondrelation_findobjfromobjsrc:
 * @bond: Global bond object.
 * 
 * Looks for a dbobjectsrc defined with the name of @dbobjectsrc.
 * 
 * Returns: DbObject stored in @dbobjectsrc if its availiable
 */
DbObject *
bondrelation_findobjfromobjsrc(Bond * bond, gchar * dbobjectsrc)
	{
	OperationsNode *oprnode;
	BondObject *bobj;
	DbObject *obj = NULL;

	if (dbobjectsrc == NULL)
		{
		errormsg("No loadby string");
		return NULL;
		}
	oprnode = g_hash_table_lookup(bond->hashdbobjectsrc, dbobjectsrc);
	if (oprnode == NULL)
		{
		warningmsg("Unable to find dbobjectsrc %s widget", dbobjectsrc);
		return NULL;
		}
	bobj = bond_getbyname(bond, oprnode->widgetname);
	if (bobj == NULL)
		{
		warningmsg("dbobjectsrc %s refers to widget %s but it hasn't been "
		           "displayed yet", dbobjectsrc, oprnode->widgetname);
		return NULL;
		}
	obj = bobj->obj;
	return obj;
	}

static gboolean
bondrelation_addobj(BondRelation * brel, DbObject * obj)
	{
	GList *walk;

	for (walk = g_list_first(brel->parentobj); walk != NULL; walk = walk->next)
		{
		if (walk->data == obj)
			{
			warningmsg("object already there. can't add again");
			return FALSE;
			}
		}
	brel->parentobj = g_list_append(brel->parentobj, obj);
	return TRUE;
	}

/**
 * bondrelation_addpending_brel:
 * @bond: Global bond object.
 * @brel: BondRelation that you've defined that you want attached at some point.
 * 
 * You can add @brel (BondRelation) to be loaded before the a widget is opened.
 * This is so that you can specify how a database object will be displayed in advance
 * of it being opened (ie Loadby use this a lot).
 * 
 * Returns: Zero. 
 */
gint
bondrelation_addpending_brel(Bond * bond, BondRelation * brel)
	{
	g_assert(brel);
	bond->brelinwait = g_list_prepend(bond->brelinwait, brel);
	return 0;
	}

/**
 * bondrelation_addpending:
 * @bond: Global bond object
 * @widgetname: 
 * @loadby: 
 * 
 * Pending is for dealing with forms not yet displayed or spawned yet. It is
 * saying in advance that this object must be loaded by this other object.
 * 
 * Returns: non-zero on error.
 */
gint
bondrelation_addpending(Bond * bond, gchar * widgetname, gchar * loadby)
	{
	GList *walk, *loadbywalk;
	BondRelation *brel;

	g_assert(bond);
	g_assert(widgetname);
	g_assert(loadby);

	for (walk = g_list_first(bond->brelinwait); walk != NULL; walk = walk->next)
		{
		brel = walk->data;
		if (strcmp(brel->targetwidget, widgetname) == 0)
			{
			for (loadbywalk = g_list_first(brel->loadbywidget);
			        loadbywalk != NULL; loadbywalk = loadbywalk->next)
				{
				if (strcmp(loadbywalk->data, loadby) == 0)
					{
					warningmsg
					("Loadby %s on %s already exists. can't add.",
					 loadby, widgetname);
					return -1;
					}
				}
			brel->loadbywidget = g_list_append(brel->loadbywidget, loadby);
			return 0;
			}
		}
	brel = bondrelation_create(RelationAuto);
	brel->loadbywidget = g_list_append(NULL, loadby);
	brel->targetwidget = widgetname;
	return bondrelation_addpending_brel(bond, brel);
	}

/**
 * bondrelation_checkpending:
 * @bond:
 * @widgetname:
 * @orginalbrel:
 * 
 * 
 * 
 * 
 * Returns: 
 */
BondRelation *
bondrelation_checkpending(Bond * bond, gchar * widgetname,
                          BondRelation * orginalbrel)
	{
	BondObject *bobjloadby;
	BondRelation *tmpbrel, *brel = NULL;
	gchar *loadbyname;
	GList *walk, *widgetwalk;

	g_assert(bond);

	if (!bond->brelinwait)
		return orginalbrel;

	for (widgetwalk = g_list_first(bond->brelinwait); widgetwalk != NULL;
	        widgetwalk = widgetwalk->next)
		{
		tmpbrel = widgetwalk->data;
		if (tmpbrel->targetwidget)
			if (strcmp(tmpbrel->targetwidget, widgetname) == 0)
				{
				if (orginalbrel)
					{
					debuglog(80,
					         "BondRelation already exists on %s, freeing it.",
					         widgetname);
					mem_free(orginalbrel);
					}
				brel = tmpbrel;
				bond->brelinwait = g_list_remove(bond->brelinwait, brel);
				/* add loadbys */
				for (walk = g_list_first(brel->loadbywidget); walk != NULL;
				        walk = walk->next)
					{
					loadbyname = walk->data;
					bobjloadby = bond_getbyname(bond, loadbyname);
					if (!bobjloadby)
						{
						errormsg
						("Unable to find bondobject for loadby %s for"
						 " showing %s", loadbyname, widgetname);
						continue;
						}
					if (!bobjloadby->obj)
						{
						errormsg
						("No database object assoacited with %s for showing %s,"
						 " its suppose to be a loadby", loadbyname,
						 widgetname);
						continue;
						}
					bondrelation_addobj(brel, bobjloadby->obj);
					}
				break;
				}
		}
	if (brel == NULL)
		return orginalbrel;
	debuglog(70, "Found brelinwait to loadby on %s", brel->targetwidget);
	return brel;
	}

/**
 * bondrelation_removepending
 * @bond: Global bond object.
 * @obj: Database object to remove pending.
 * @bobj:
 *
 * Checks the brelinwait for anything with a DbObject stored the same as @obj, if it is so it will
 * remove this bondrelation in wait.  This is to stop courption of the @obj stored in
 * the bondrelation.
 */

/* Tempoarlly disabled because i see no point.. why did i code this?? brelinwait only includes
 * widgetnames not dbobjs */
#if 0
GList *
bondrelation_removepending(Bond * bond, DbObject * obj)
	{
	GList *walk, *objwalk, *tofree = NULL;
	gboolean free;
	BondRelation *brel;

	for (walk = g_list_first(bond->brelinwait); walk != NULL; walk = walk->next)
		{
		brel = walk->data;
		free = false;

		if (brel->bobj == bobj)
			{
			tofree = g_list_prepend(tofree, brel);
			free = TRUE;
			}

		for (walk = g_list_first(brel->parentobj); walk != NULL;
		        walk = walk->next)
			{
			if (walk->data == obj)
				{
				tofree = g_list_prepend(tofree, brel);
				free = TRUE;
				break;
				}
			}
		if (tofree)
			{
			if (brel->bobj)
				debuglog(60, "Removing item %s from brelinwaiting because it "
				         "depends on a dbobject that has been freed",
				         brel->bobj->widgetname);
			else
				debuglog(60, "Removing an item from the brelinwating list.");
			bond->brelinwait = g_list_remove(bond->brelinwait, brel);
			continue;
			}
		}
	return tofree;
	}
#endif

/**
 * bondrelation_anaylise:
 * @bobj: Bond object to load database objects or set fields for.
 *
 * Work out what needs to be done to load the data for this
 * @bobj and then load the data if any dbobject is specified
 * else find the table/field if it is assocated with one.
 * 
 * 
 * Returns: non-zero on error.
 */
gint
bondrelation_anaylise(BondObject * bobj)
	{
	gint retval = -1;
	DbObject *newobj = NULL, *oldobj;
	BondObject *tmpbobj;
	OperationsNodeDb *dbobj;

	g_assert(bobj);
	g_assert(bobj->oprnode);

	/* See if its a datasource, and if it is load it up */
	dbobj = bobj->oprnode->dbobj;
	if (dbobj != NULL)
		{
		/* m.essage("Activating %s", bobj->widgetname); */
		bobj->dbobjactive = TRUE;
		/* check to see if there are any brel in wait */
		bobj->brel = bondrelation_checkpending(bobj->bond, 
					bobj->widgetname, 
		                        bobj->brel);
		/* If a bondrelationship doesn't exist, create a auto one */
		if (bobj->brel == NULL)
			bobj->brel = bondrelation_create(RelationAuto);
		bobj->brel->dbobj = dbobj;
		oldobj = bobj->obj;
		bobj->brel->bobj = bobj;
		/* Load relevant data, if it needs to be loaded */
		/* Note, NULL is passed as obj so that old value can be 
		 * retained for checking */
		if (dbobj->dontexec == TRUE)
			{
			if (oldobj)
				newobj = oldobj;
			else
				newobj = db_createobject(NULL);
			}
		else
		if ((newobj = bondrelation_checkcache(bobj)) == NULL)
			newobj = bondrelation_load(bobj->bond, NULL, bobj->brel);
		/* change any references to this object to the new object */
		/* Delete this and line above */
		/* bondrelation_changeobjs(bobj->bond, newobj, oldobj); */
		/* free the old object */
		/* remove comments here - mmeory leak */
		/* if (oldobj) db_obj_debug(oldobj); message("========== output from
		   analyis - newobj"); db_obj_debug(newobj); message("==========
		   output from analyis - done"); */
		if (oldobj != newobj)
			bondrelation_changeobjs(bobj->bond, newobj, oldobj);

		if (oldobj && newobj)
			{
			db_freeobject(oldobj);
			}
		if (newobj)
			{
			/* this would already been set if oldobj not null */
			bobj->obj = newobj;
			/* Apply the nesserary fitlers to the database object */
			bondfilter_applyfilters(bobj);
			if (db_numrecord(bobj->obj) > 0)
				db_moveto(bobj->obj, 0);
			retval = 0;
			}
		else 
			{
			warningmsg("No bondobject was loaded.");
			retval = -1;
			}
		}
	/* This is for working out the field type */
	switch (bobj->oprnode->widgettype)
		{
	case UiEntryBox:
	case UiGrid:
	case UiDropDown:
	case UiToggleButton:
	case UiCheckButton:
	case UiSpinButton:
	case UiRadioButton:
	case UiTextBox:
		bobj->active = TRUE;
		break;
	default:
		break;
		}
	/* work out if its a field, or a something else */
	if (bobj->oprnode->field != NULL || bobj->widgettype == UiGrid)
		{
		if (bobj->oprnode->field != NULL && bobj->widgettype != UiGridColumn)
			bobj->fieldactive = TRUE;
		/* find its parent database obj */
		if (bobj->oprnode->dbobjectsrc != NULL)
			{
			if (bobj->obj == NULL)
				{
				g_assert(bobj->oprnode->dbobjectsrc_widget);
				tmpbobj = bond_getbyname(bobj->bond,
				                         bobj->oprnode->
				                         dbobjectsrc_widget);

				retval = 0;
				if (tmpbobj == NULL)
					{
					errormsg("Unable to find bond object. This means i"
					         " can't locate a live database object for"
					         " the desired dbobjectsrc %s, widget %s, in %s",
					         bobj->oprnode->dbobjectsrc,
					         bobj->oprnode->dbobjectsrc_widget,
						 bobj->widgetname);
					debugmsg
					("This could be caused by the order of the xmlfile,"
					 " make sure you reference a"
					 " dbobjectsrc that has already been passed.");
					retval = -2;
					g_assert(NULL);
					}
				else
					{
					bobj->obj = tmpbobj->obj;
					if (tmpbobj->obj == NULL)
						warningmsg("... obj is null from %s",
						           tmpbobj->widgetname);
					bobj->dbobjectsrc = tmpbobj;
					}
				}
			}
		else if (bobj->widgettype != UiGrid)
			warningmsg("No dbobjectsrc defined on widget %s", bobj->widgetname);

		if (bobj->obj == NULL)
			warningmsg("Can not find a database object for widget %s but"
			           " it has field %s defined. dbobjectsrc is %s and its"
			           " presumed widget is %s", bobj->widgetname,
			           bobj->oprnode->field, bobj->oprnode->dbobjectsrc,
			           bobj->oprnode->dbobjectsrc_widget);
		}

	return retval;
	}

/**
 * bondrelation_refresh:
 * @bobj: Bond Object
 * 
 * 
 * 
 */
gint
bondrelation_refresh(BondObject * bobj)
	{
	gint oldrow = -1;
	DbObject *oldobj, *newobj;

	g_assert(bobj);
	g_assert(bobj->brel);
	/* get old position */
	if (bobj->obj)
		{
		oldrow = db_currentrow(bobj->obj);
		/* db_clearobject(bobj->obj); */
		}

	bobj->brel = bondrelation_checkpending(bobj->bond, bobj->widgetname,
	                                       bobj->brel);
	g_assert(bobj->obj);
	/* if (!bobj->obj || db_isdeadobject(bobj->obj) == TRUE) { m.essage("Doing
	   apply thing"); bondrelation_anaylise(bobj); m.essage("-- done anaylis");
	   return 1; } */
	/* load the database object again. Run the search if thats required */
	if ((bondrelation_checkcache(bobj)) == NULL)
		{
		/* m.essage("Do bondrelation load thing. %s",bobj->obj->basetable); */
		oldobj = bobj->obj;
		newobj = bondrelation_load(bobj->bond, bobj->obj, bobj->brel);
		/* change any references to this object to the new object */
		if (newobj != oldobj)
			{
			warningmsg("this shouldn't happen?");
			bondrelation_changeobjs(bobj->bond, newobj, oldobj);
			}
		/* m.essage("newobj is %s or
		   %s",bobj->obj->basetable,newobj->basetable); */
		bobj->obj = newobj;
		}
	if (!bobj->obj)
		{
		debuglog(95, "Obj is null, so wont apply filter to it");
		return -2;
		}
	/* apply any filters */
	bondfilter_applyfilters(bobj);
	/* move back to old row if its possible */
	if (oldrow >= 0 && oldrow < db_numrecord(bobj->obj))
		{
		debuglog(70, "Refreshing BondObject %s and moving back to "
			"orginal position %d", bobj->widgetname, oldrow);
		db_moveto(bobj->obj, oldrow);
		}
	else
		debuglog(70, "Refreshing BondObject %s done", bobj->widgetname);
	return 0;
	}

/**
 * bondrelation_saveeffected:
 * @bond: Global bond object.
 * @objlist: 
 * 
 * Given a list of database objects, save any widget that references these
 * database objects.
 * 
 * Returns: Non-zero on error.
 */
gint
bondrelation_saveeffected(Bond * bond, GList * objlist)
	{
	GList *walk, *effectedwalk;
	GList *effected;
	DbObject *obj;
	BondObject *bobj;

	for (walk = g_list_first(objlist); walk != NULL; walk = walk->next)
		{
		obj = walk->data;
		/* Save any bond object forms that contain that obj */
		effected = g_hash_table_lookup(bond->globalbobj_byobj, obj);
		debuglog(70, "%d items effected by %s. saving them now",
		         g_list_length(effected), obj->basetable);
		for (effectedwalk = g_list_first(effected); effectedwalk != NULL;

		        effectedwalk = effectedwalk->next)
			{
			bobj = effectedwalk->data;
			bond_save(bobj, TRUE, FALSE);
			}
		}

	return 0;
	}

/**
 * Handling dependon in regard to watch registrations.
 * 
 * Wait till spawn_rec to finish.
 * Recusively go though bond objects, if have field defined.
 * If tablename isn't the basetable of object, set dependon list.
 * Dependonlist is generated 
 */
