/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <strings.h>
#include <stdlib.h>

#include <glib.h>
#include <libxml/parser.h>

#include "db.h"
#include "dbsqlparse.h"
#include "bc.h"
#include "xml_funcs.h"
#include "sql_parser.h"

#include "bond.h"
#include "bondxml.h"
#include "bondmain.h"
#include "fieldpath.h"

/**
 * bondxml_init
 * @bond: Bond structure
 *
 * bondxml_init() initialises the bond structure for parsing.
 *
 * Returns: Result of a function call to bondxml_parse()
 */
gint
bondxml_init(Bond * bond)
	{
	xmlNode *node;

	g_assert(bond);
	g_assert(bond->uixml);
	g_assert(bond->uixml->doc);

	/* Get root element from @bond's uixml document tree */
	node = xmlDocGetRootElement(bond->uixml->doc);
	if (!IsEqual(node->name, "bond"))
		errormsg("Root element should be bond\n");

	debuglog(100, "Root node name is: %s\n", node->name);
	if (!node)
		errormsg("Can't get root element from xml document tree.");

	g_assert(!bond->hashoprnode);

	/* Initialise new hash tables for operations nodes and db object src */
	bond->hashoprnode =
	    g_hash_table_new_full(g_str_hash, g_str_equal, NULL, bondxml_free);
	bond->hashdbobjectsrc =
	    g_hash_table_new_full(g_str_hash, g_str_equal, NULL, NULL);
	/* This starts at the <window> tag and works it way down the list. */
	/* Return result of parsing the bondxml doc tree */
	return bondxml_parse(bond, node, NULL);
	}

/*
 * bondxml_cleanup:
 * @bond: Bond structure
 *
 * bondxml_cleanup() has not been written yet.
 *
 * Returns: -1
 */
gint
bondxml_cleanup(Bond * bond)
	{
	g_assert(bond);
	g_hash_table_destroy(bond->hashoprnode);
	g_hash_table_destroy(bond->hashdbobjectsrc);

	/* ... . . Now I sure i forgot to free something */
	xmlFreeDoc(bond->uixml->doc);
	/* what else? */
	return -1;
	}

/**
 * bondxml_free:
 * @ptr: Pointer to what is hopefully an operations node.
 *
 * bondxml_free() frees all memory used by *ptr.
 *
 * Returns: void
 */
void
bondxml_free(void *ptr)
	{
	OperationsNode *oprnode = (OperationsNode *) ptr;
	OperationsNodeTrigger *oprtrigger;
	OperationsNodeTriggerArgument *tmpargument;
	GList *walk, *innerwalk;

	/* Free the dbobject component and its children */
	if (oprnode->dbobj)
		{
		/* free a filters list */
		for (walk = g_list_first(oprnode->dbobj->filters);
		        walk != NULL; walk = walk->next)
			{
			g_assert(walk->data);
			mem_free(walk->data);
			}
		if (oprnode->dbobj->filters)
			g_list_free(oprnode->dbobj->filters);
		/* free search list */
		for (walk = g_list_first(oprnode->dbobj->search);
		        walk != NULL; walk = walk->next)
			mem_free(walk->data);
		if (oprnode->dbobj->search)
			g_list_free(oprnode->dbobj->search);
		/* free rest of dbobject */
		mem_free(oprnode->dbobj);
		}

	/* free any method's/triggers defined on the widget */
	if (oprnode->triggerlist)
		{
		for (walk = g_list_first(oprnode->triggerlist);
		        walk != NULL; walk = walk->next)
			{
			oprtrigger = walk->data;
			/* free the argument list */
			for (innerwalk = g_list_first(oprtrigger->argumentlist);
			        innerwalk != NULL; innerwalk = innerwalk->next)
				{
				tmpargument = innerwalk->data;
				if (tmpargument->fieldpathlist)
					fieldpath_free(
						tmpargument->fieldpathlist);
				mem_free(tmpargument);
				}

			if (oprtrigger->argumentlist)
				g_list_free(oprtrigger->argumentlist);
			/* free any loadby's specified. (String stored in 
			 * xmldoc) */
			if (oprtrigger->loadby)
				g_list_free(oprtrigger->loadby);
			mem_free(oprtrigger);
			}
		g_list_free(oprnode->triggerlist);
		}

	if (oprnode->fieldlist)
		g_list_free(oprnode->fieldlist);
	if (oprnode->tablelist)
		g_list_free(oprnode->tablelist);
	if (oprnode->variablestofree)
		{
		for (walk = g_list_first(oprnode->variablestofree); walk;
		        walk = walk->next)
			mem_free(walk->data);
		g_list_free(oprnode->variablestofree);
		}
	/* Free the populatelist used in dropdown boxes, including each 
	 * element */
	for (walk = g_list_first(oprnode->populatelist); walk; 
			walk = walk->next)
		g_free(walk->data);

	if (oprnode->populatelist)
		g_list_free(oprnode->populatelist);
	/* Finish what was started */
	mem_free(oprnode);
	}

/**
 * bondxml_getnode:
 * @bond: Bond structure
 * @widgetname: Name of widget
 *
 * Gets operations node with the same name as widgetname.
 *
 * Returns: Operations node with name widgetname
 */
OperationsNode *
bondxml_getnode(Bond * bond, gchar * widgetname)
	{
	return g_hash_table_lookup(bond->hashoprnode, widgetname);
	}

/*
 * bondxml_parsewidget_createfilter:
 * @node: A node in the xml tree
 * @widgetname: Name of widget
 *
 * Creates an operations node filter from the widgetname and xml node. 
 * 
 * Returns: Operations node filter
 */
static OperationsNodeFilter *
bondxml_parsewidget_createfilter(xmlNode * node, gchar * widgetname)
	{
	OperationsNodeFilter *tmpfilter;
	gchar *content;

	/* Create an operations node filter */
	tmpfilter =
	    (OperationsNodeFilter *) mem_calloc(sizeof(OperationsNodeFilter));
	/* Assign values same as node and widgetname */
	tmpfilter->widgetname = widgetname;
	tmpfilter->field = xmlGetProp(node, "field");
	tmpfilter->table = xmlGetProp(node, "table");

	/* Assign content the value of "opr" from the inputted node. Analyze 
	 * and give an SQL token value to tmpfilter->opr which corresponds 
	 * to the value of content. */
	content = xmlGetProp(node, "opr");

	if (content)
		{
		if (IsEqual(content, "equal"))
			tmpfilter->opr = SQL_eq;
		else if (IsEqual(content, "=="))
			tmpfilter->opr = SQL_eq;
		else if (IsEqual(content, "="))
			tmpfilter->opr = SQL_eq;
		else if (IsEqual(content, "like"))
			tmpfilter->opr = SQL_like;
		else if (IsEqual(content, "notequal"))
			tmpfilter->opr = SQL_not;
		else if (IsEqual(content, "!equal"))
			tmpfilter->opr = SQL_not;
		else if (IsEqual(content, "!="))
			tmpfilter->opr = SQL_not;
		else if (IsEqual(content, "greater"))
			tmpfilter->opr = SQL_gt;
		else if (IsEqual(content, ">"))
			tmpfilter->opr = SQL_gt;
		else if (IsEqual(content, "<"))
			tmpfilter->opr = SQL_lt;
		else if (IsEqual(content, "less"))
			tmpfilter->opr = SQL_lt;
		else if (IsEqual(content, "<="))
			tmpfilter->opr = SQL_leq;
		else if (IsEqual(content, "lessequal"))
			tmpfilter->opr = SQL_leq;
		else if (IsEqual(content, ">="))
			tmpfilter->opr = SQL_geq;
		else if (IsEqual(content, "greaterequal"))
			tmpfilter->opr = SQL_geq;
		}

	return tmpfilter;
	}

/**
 * bondxml_parsewidget_createsearch:
 * @node: A node in the xml tree
 *
 * Similar to bondxml_parsewidget_createfilter(), creating a search
 * rather than a filter.
 * 
 * Returns: An operations node search
 */
static OperationsNodeSearch *
bondxml_parsewidget_createsearch(xmlNode * node)
	{
	OperationsNodeSearch *tmpsearch;
	gchar *content;

	/* Create an operations node search */
	tmpsearch =
	    (OperationsNodeSearch *) mem_calloc(sizeof(OperationsNodeSearch));
	/* Similar to bondxml_parsewidget_createfilter() */
	tmpsearch->field = xmlGetProp(node, "field");
	tmpsearch->table = xmlGetProp(node, "table");
	content = xmlGetProp(node, "opr");
	if (content)
		{
		if (IsEqual(content, "equal"))
			tmpsearch->opr = SQL_eq;
		else if (IsEqual(content, "=="))
			tmpsearch->opr = SQL_eq;
		else if (IsEqual(content, "="))
			tmpsearch->opr = SQL_eq;
		else if (IsEqual(content, "like"))
			tmpsearch->opr = SQL_like;
		else if (IsEqual(content, "notequal"))
			tmpsearch->opr = SQL_not;
		else if (IsEqual(content, "!equal"))
			tmpsearch->opr = SQL_not;
		else if (IsEqual(content, "!="))
			tmpsearch->opr = SQL_not;
		else if (IsEqual(content, "greater"))
			tmpsearch->opr = SQL_gt;
		else if (IsEqual(content, ">"))
			tmpsearch->opr = SQL_gt;
		else if (IsEqual(content, "<"))
			tmpsearch->opr = SQL_lt;
		else if (IsEqual(content, "less"))
			tmpsearch->opr = SQL_lt;
		else if (IsEqual(content, "<="))
			tmpsearch->opr = SQL_leq;
		else if (IsEqual(content, "lessequal"))
			tmpsearch->opr = SQL_leq;
		else if (IsEqual(content, ">="))
			tmpsearch->opr = SQL_geq;
		else if (IsEqual(content, "greaterequal"))
			tmpsearch->opr = SQL_geq;
		}
	/* Check case sensitivity and modify tmpsearch is it is set */
	content = xmlGetProp(node, "casesensitive");
	if (content && IsTrue(content))
		tmpsearch->casesensitive = TRUE;

	return tmpsearch;
	}

/**
 * bondxml_parsewidget_createargument:
 * @node: A node in the xml tree
 *
 * Create an argument with the name of the node.
 * 
 * Returns: An operations node trigger argument
 */
static OperationsNodeTriggerArgument *
bondxml_parsewidget_createargument(xmlNode * node)
	{
	OperationsNodeTriggerArgument *argument;

	argument = mem_calloc(sizeof(OperationsNodeTriggerArgument));
	argument->name = xmlGetProp(node, "name");

	return argument;
	}

/**
 * bondxml_parsewidget_createtrigger:
 * @node: A node in the xml tree
 * 
 * Creates an operations node trigger with inputted node's event.
 *
 * Returns: An operations node trigger
 */
static OperationsNodeTrigger *
bondxml_parsewidget_createtrigger(xmlNode * node)
	{
	gchar *content;
	OperationsNodeTrigger *trigger;

	/* Create an operations node trigger */
	trigger = mem_calloc(sizeof(OperationsNodeTrigger));
	trigger->event = EventNone;

	content = xmlGetProp(node, "event");
	if (!content || content[0] == 0)
		{
		warningmsg("No event is specified on trigger at line %d",
		           xmlGetLineNo(node));
		return trigger;
		}

	if (IsEqual(content, "default"))
		trigger->event = EventDefault;
	else if (IsEqual(content, "change"))
		trigger->event = EventChange;
	else if (IsEqual(content, "entry"))
		trigger->event = EventEntry;
	else if (IsEqual(content, "exit"))
		trigger->event = EventExit;
	else if (IsEqual(content, "click"))
		trigger->event = EventClick;
	else if (IsEqual(content, "beforepaint"))
		trigger->event = EventBeforePaint;
	else if (IsEqual(content, "afterpaint"))
		trigger->event = EventAfterPaint;
	else if (IsEqual(content, "beforesave"))
		trigger->event = EventBeforeSave;
	else if (IsEqual(content, "aftersave"))
		trigger->event = EventAfterSave;
	else if (IsEqual(content, "select"))
		trigger->event = EventSelect;
	else if (IsEqual(content, "open"))
		trigger->event = EventOpen;
	else if (IsEqual(content, "close"))
		trigger->event = EventClose;
	else if (IsEqual(content, "add"))
		trigger->event = EventAdd;
	else if (IsEqual(content, "delete"))
		trigger->event = EventDelete;
	else if (IsEqual(content, "gridedit"))
		trigger->event = EventGridEdit;
	/* else if (IsEqual(content, "sql")) */

	/* trigger->event = EventSQL; */
	else
		warningmsg("Invalid event specified in trigger at line %d",
		           xmlGetLineNo(node));

	return trigger;
	}

static ApiCall
bondxml_parsewidget_createapi(xmlNode * node, OperationsNodeTrigger *trigger)
	{
	gchar *content;
	gchar *confirm;

	g_assert(trigger);
	trigger->apicall = ApiNone;
	trigger->confirmaction = FALSE;
	confirm = xmlGetProp(node, "confirm");
	if (confirm && confirm[0] != 0)
		if (IsEqual(confirm, "true"))
			trigger->confirmaction = TRUE;
	
	content = xmlGetProp(node, "call");
	if (!content || content[0] == 0)
		{
		warningmsg("No call specified attribute found on api call on "
			"line %d", xmlGetLineNo(node));
		return trigger->apicall;
		}
		
	if (IsEqual(content, "open"))
		trigger->apicall =  ApiOpen;
	else if (IsEqual(content, "close"))
		trigger->apicall =  ApiClose;
	else if (IsEqual(content, "add"))
		trigger->apicall =  ApiAdd;
	else if (IsEqual(content, "new"))
		trigger->apicall =  ApiNew;
	else if (IsEqual(content, "delete"))
		trigger->apicall =  ApiDelete;
	else if (IsEqual(content, "loadby"))
		trigger->apicall =  ApiLoadBy;
	else if (IsEqual(content, "loadall"))
		trigger->apicall =  ApiLoadAll;
	else if (IsEqual(content, "sql"))
		trigger->apicall =  ApiSQL;
	else if (IsEqual(content, "refresh"))
		trigger->apicall =  ApiRefresh;
	else if (IsEqual(content, "save"))
		trigger->apicall =  ApiSave;
	else if (IsEqual(content, "abortsave"))
		trigger->apicall =  ApiAbortSave;
	else if (IsEqual(content, "clear"))
		trigger->apicall =  ApiClear;
	else if (IsEqual(content, "flush"))
		trigger->apicall =  ApiFlush;
	else if (IsEqual(content, "quit"))
		trigger->apicall =  ApiQuit;
	else if (IsEqual(content, "report"))
		trigger->apicall =  ApiReport;
	else if (IsEqual(content, "next"))
		trigger->apicall =  ApiNext;
	else if (IsEqual(content, "prev"))
		trigger->apicall =  ApiPrev;
	else if (IsEqual(content, "first"))
		trigger->apicall =  ApiFirst;
	else if (IsEqual(content, "last"))
		trigger->apicall =  ApiLast;
	else if (IsEqual(content, "moveto"))
		trigger->apicall =  ApiMoveTo;
	else
		warningmsg("Invalid API call specified at line %d.", 
				xmlGetLineNo(node));
	return trigger->apicall;
	}

static void
bondxml_parsewidget_populatelist(Bond * bond, xmlNodePtr node,
                                 OperationsNode * oprnode)
	{
	gchar *content;

	content = xmlGetProp(node, "saveoffset");
	if (content && IsTrue(content))
		oprnode->saveoffset = TRUE;

	oprnode->populatelist = ui_dropdown_getpopulatelist(bond->uixml, node);
	}

static OperationsNode *
bondxml_getnodebyobjectsrc(Bond * bond, gchar * dbobjectsrc)
	{
	return g_hash_table_lookup(bond->hashdbobjectsrc, dbobjectsrc);
	}

static gint
bondxml_getdbobjectsrc(OperationsNode * oprnode)
	{
	if (oprnode->parentnode)
		{
		oprnode->dbobjectsrc = oprnode->parentnode->dbobjectsrc;
		oprnode->dbobjectsrc_widget = 
			oprnode->parentnode->dbobjectsrc_widget;
		/* m.essage("Setting %s in
		   %s",oprnode->parentnode->dbobjectsrc_widget,
		   oprnode->parentnode->dbobjectsrc); */ 
		return 0;
		}
	return -1;
	}

static gint
bondxml_parsewidget_findfieldpath(OperationsNode * oprnode,
                                  OperationsNodeTriggerArgument * tmpargument)
	{
	if (tmpargument->table && tmpargument->field && oprnode->field
	        && oprnode->table)
		{
		tmpargument->fieldpathlist =
		    fieldpath_init(oprnode->table, oprnode->field, 
				    tmpargument->table);
		return 0;
		}
	else if (tmpargument->table && tmpargument->field)
		tmpargument->fieldpathtomake = TRUE;
	return 0;
	}

/* Add information created from the database */
static gint
bondxml_addfieldinfo(Bond * bond, OperationsNode * oprnode)
	{
	OperationsNode *tmpnode;

	/* return if has no database source, else set it */
	if (!oprnode->dbobjectsrc)
		/* search up the tree finding a dbobjectsrc somewhere */
		if (bondxml_getdbobjectsrc(oprnode) < 0)
			return 2;
	if (oprnode->field && oprnode->dbobjectsrc)
		{
		if (!oprnode->table)
			{
			tmpnode = bondxml_getnodebyobjectsrc(bond, 
					oprnode->dbobjectsrc);
			if (!tmpnode)
				{
				warningmsg("Unable to find the dbobjectsrc "
				"%s specified in dbobjectsrc %s specified.",
				 oprnode->dbobjectsrc, oprnode->field);
				return -1;
				}
			if (!tmpnode->dbobj)
				return -2;
			/* get the table */
			oprnode->table = tmpnode->dbobj->table;
			}

		return 0;
		}
	else
		{
		oprnode->field = NULL;
		oprnode->table = NULL;
		}
	return 1;
	}

/* Generate a name for a widget, used for dropdown lookup names */
static gchar *
bondxml_getwidgetname(xmlNode * node, gchar * parentwidgetname)
	{
	gchar *retstr = NULL, *longname;
	static gint counter = 0;

	g_assert(parentwidgetname);
	longname =
	    mem_strdup_printf("%s/%s%d", parentwidgetname, node->name, counter);
	xmlNewProp(node, "name", longname);
	mem_free(longname);
	retstr = xmlGetProp(node, "name");
	counter++;
	return retstr;
	}

/* Called when you create a new database object. dbobjectsrc is set to itself */
static gint
bondxml_parsewidget_adddb(xmlNode * node, OperationsNode * oprnode)
	{
	if (oprnode->dbobj)
		{
		errormsg("Multiple declarations of dbobject in %s",
		         oprnode->widgetname);
		return -1;
		}
	oprnode->dbobj = (OperationsNodeDb *) mem_calloc(sizeof(OperationsNodeDb));

	oprnode->dbobj->name = xmlGetProp(node, "name");
	if (!oprnode->dbobj->name)
		oprnode->dbobj->name = oprnode->widgetname;

	oprnode->dbobj->widgetname = oprnode->widgetname;
	oprnode->dbobjectsrc = oprnode->dbobj->name;
	oprnode->dbobjectsrc_widget = oprnode->widgetname;

	/* Sanity check for bondxml doc. A dbobject must occur in a standard 
	 * widget or in a lookup. It wont work in table cells, other dbobjects,
	 * or fields etc */
	if (!oprnode->widget && !IsEqual(oprnode->nodename, "lookup"))
		{
		warningmsg("OperationNode->Widget is null on %s at %s. "
		           "This will cause any field that references this "
		           "dbobjectsrc to fail horribly with unable to find "  
		           "widget. Please move the dbobject xml defination to "     		           "widget defination", oprnode->dbobj->name, 
			   oprnode->widgetname);
		}
	/* m.essage("Adding dbobjectsrc %s on %s", oprnode->dbobjectsrc,
	   oprnode->widgetname); */ 
	return 0;
	}

/*
 * bondxml_create:
 * @bond: Bond structure
 * @node: Current node in the xml parsing tree
 * @widgetname: Name of the current widget
 *
 * Create an operations node.
 *
 * Returns: The created operations node.
 */
static OperationsNode *
bondxml_create(Bond * bond, xmlNode * node, gchar * widgetname)
	{
	OperationsNode *oprnode;

	oprnode = (OperationsNode *) mem_calloc(sizeof(OperationsNode));
	oprnode->name = xmlGetProp(node, "name");
	oprnode->widgetname = widgetname;
	oprnode->nodename = (gchar *) node->name;
	if (strlen(widgetname) > 1)
		oprnode->widget = ui_getwidget(bond->uixml, widgetname);
	oprnode->allowednullvalue = TRUE;
	if (oprnode->widget)
		{
		oprnode->widgettype = ui_gettype(bond->uixml, oprnode->widget);
		oprnode->defaultvisible = 
			ui_isvisible(bond->uixml, oprnode->widget);
		}
	else
		{
		oprnode->widgettype = UiNull;
		oprnode->defaultvisible = TRUE;
		}
	return oprnode;
	}

/*
 * bondxml_parsewidget:
 * @bond: Bond structure
 * @node: Node in the xml parsing tree
 * @state: Current parse state
 * @widgetname: Name of current widget
 * @oprnode: Current operations node
 * @parentnode: Parent node
 * @trigger: Current operations node trigger
 *
 * Recursive function for parsing a widget.  One oh mightly long C function.  You
 * know how norty i can be.
 *
 * Returns: If successful, an operations node. Else NULL.
 */
static OperationsNode *
bondxml_parsewidget(Bond * bond, xmlNode * node, ParseState state,
                    gchar * widgetname, OperationsNode * oprnode,
                    OperationsNode * parentnode,
                    OperationsNodeTrigger * trigger)
	{
	gchar *lookupname, *tmptable, *content, *nodename;
	gboolean addtohash = FALSE, multifield = FALSE;
	OperationsNode *tmpnode;
	OperationsNodeFilter *tmpfilter = NULL;
	OperationsNodeSearch *tmpsearch = NULL;
	OperationsNodeTrigger *tmptrigger = NULL;
	OperationsNodeTriggerArgument *tmpargument = NULL;

	g_assert(node);
	g_assert(widgetname);

	if (!oprnode)
		{
		/* Work out what type of widget it is, create the operations 
		 * node for that widget */
		oprnode = bondxml_create(bond, node, widgetname);
		oprnode->parentnode = parentnode;
		/* If there exists a @parentnode, add current operations node 
		 * to parents list of children nodes */
		if (parentnode)
			parentnode->childrennode =
			    g_list_append(parentnode->childrennode, oprnode);
		addtohash = TRUE;
		}
	/* create filter/search/trigger etc if you need to */
	switch (state)
		{
	case ParseFilter:
		tmpfilter = bondxml_parsewidget_createfilter(node, widgetname);
		/* If there doesn't exist a table in the filter, create one 
		 * from oprnode->dbobj->table (if it exists). */
		if (!tmpfilter->table && oprnode && oprnode->dbobj)
			tmpfilter->table = oprnode->dbobj->table;
		oprnode->dbobj->filters =
		    g_list_append(oprnode->dbobj->filters, tmpfilter);
		break;
	case ParseSearch:
		tmpsearch = bondxml_parsewidget_createsearch(node);
		if (!tmpsearch->table && oprnode && oprnode->dbobj)
			tmpsearch->table = oprnode->dbobj->table;
		oprnode->dbobj->search =
		    g_list_append(oprnode->dbobj->search, tmpsearch);
		break;
	case ParseTrigger:
		tmptrigger = bondxml_parsewidget_createtrigger(node);
		if (!tmptrigger)
			return NULL;
		tmptrigger->bond = bond;
		oprnode->triggerlist =
		    g_list_prepend(oprnode->triggerlist, tmptrigger);
		break;
	case ParseArgument:
		if (!trigger)
			{
			errormsg("Trigger is NULL");
			return NULL;
			}
		tmpargument = bondxml_parsewidget_createargument(node);
		if (!tmpargument)
			return NULL;

		trigger->argumentlist =
		    g_list_prepend(trigger->argumentlist, tmpargument);
		break;
	case ParseAPI:
		if (!trigger)
			{
			errormsg("Trigger is NULL");
			return NULL;
			}
		bondxml_parsewidget_createapi(node, trigger);
		break;
	default:
		break;
		}

	/* Iterate through @node's children for this functions main loop */
	for (node = node->children; node; node = node->next)
		{
		nodename = (gchar *) node->name;
		switch (state)
			{
		case ParseStart:
			if (IsEqual(nodename, "dbobject"))
				{
				bondxml_parsewidget_adddb(node, oprnode);
				bondxml_parsewidget(bond, node, ParseDbObject,
				                    widgetname, oprnode, 
						    parentnode, NULL);
				break;
				}
			else if (IsEqual(nodename, "field"))
				{
				oprnode->field = g_strstrip(xmlNodeGetContent(node));
				content = xmlGetProp(node, "readonly");
				if (content && IsTrue(content))
					oprnode->readonly = TRUE;
				
				content = xmlGetProp(node, "notnull");
				if (content && IsTrue(content))
					oprnode->allowednullvalue = FALSE;
				else
					oprnode->allowednullvalue = TRUE; 
				
				tmptable = xmlGetProp(node, "table");
				if (tmptable)
					if (strlen(tmptable) > 0)
						oprnode->table = tmptable;
				break;
				}
			else if (IsEqual(nodename, "inputrestrict"))
				{
				content = xmlNodeGetContent(node);
				if (IsEqual(content, "int"))
					oprnode->inputrestrict = InputRestrictInt;
				else if (IsEqual(content, "integer"))
					oprnode->inputrestrict = InputRestrictInt;
				else if (IsEqual(content, "string"))
					oprnode->inputrestrict = InputRestrictString;
				else if (IsEqual(content, "float"))
					oprnode->inputrestrict = InputRestrictFloat;
				else if (IsEqual(content, "date"))
					oprnode->inputrestrict = InputRestrictDate;
				else if (IsEqual(content, "time"))
					oprnode->inputrestrict = InputRestrictTime;
				else if (IsEqual(content, "firsttoupper"))
					oprnode->inputrestrict = InputRestrictFirstToUpper;
				else if (IsEqual(content, "toupper"))
					oprnode->inputrestrict = InputRestrictToUpper;
				else if (IsEqual(content, "tolower"))
					oprnode->inputrestrict = InputRestrictToLower;
				else if (IsEqual(content, "phone"))
					oprnode->inputrestrict = InputRestrictPhone;
				else if (IsEqual(content, "all"))
					oprnode->inputrestrict = InputRestrictAll;
				else if (IsEqual(content, "money"))
					oprnode->inputrestrict = InputRestrictFloat;
				else if (IsEqual(content, "round"))
					oprnode->inputrestrict = InputRestrictRound;
				
				/* If we are to mask input somehow */
				/* content = xmlGetProp(node, "mask");
				 if (content && oprnode->inputrestrict == 
				 InputRestrictRound)
					oprnode->restrictmask = content; */
				break;
				}
			else if (IsEqual(nodename, "nulllistvalue"))
				{
				oprnode->nulllistvalue = xmlNodeGetContent(node);
				break;
				}
			else if (IsEqual(nodename, "dbobjectsrc"))
				{
				oprnode->dbobjectsrc = xmlNodeGetContent(node);
				break;
				}
			else if (IsEqual(nodename, "parent"))
				{
				oprnode->parent =
				    g_list_prepend(oprnode->parent,
				                   xmlNodeGetContent(node));
				break;
				}
			else if (IsEqual(nodename, "lookup"))
				{
				lookupname = bondxml_getwidgetname(node, widgetname);
				tmpnode =
				    bondxml_parsewidget(bond, node, ParseLookup,
				                        lookupname, NULL, 
							oprnode, NULL);
				break;
				}
			else if (IsEqual(nodename, "populatelist"))
				{
				bondxml_parsewidget_populatelist(bond, node, 
						oprnode);
				break;
				}
			else if (IsEqual(nodename, "trigger"))
				{
				bondxml_parsewidget(bond, node, ParseTrigger,
				                    widgetname, oprnode, 
						    parentnode, NULL);
				break;
				}
			else if (IsEqual(nodename, "nullvalue"))
				{
				oprnode->nullvalue = xmlNodeGetContent(node);
				break;
				}
			else if (IsEqual(nodename, "table"))
				{
				warningmsg("Table tag here is deprecated. Line %d",
				           xmlGetLineNo(node));
				oprnode->table = xmlNodeGetContent(node);
				break;
				}
			else if (IsEqual(nodename, "notnull"))
				{
				if (IsTrue(xmlNodeGetContent(node)))
					oprnode->allowednullvalue = FALSE;
				else
					oprnode->allowednullvalue = TRUE;
				break;
				}
			else if (IsEqual(nodename, "childwindow"))
				{
				oprnode->childrenwindow = g_list_append(
					oprnode->childrenwindow, 
					xmlNodeGetContent(node));
				}				
			break;
		case ParseDbObject:
			if (IsEqual(nodename, "table"))
				{
				if (oprnode->dbobj->sql)
					warningmsg
					("You have specified both a table and a SQL"
					 " statement. This shouldn't be done at line %d",
					 xmlGetLineNo(node));
				oprnode->dbobj->table = xmlNodeGetContent(node);
				if (!oprnode->table)
					oprnode->table = oprnode->dbobj->table;
				if (!oprnode->primary_table)
					oprnode->primary_table = oprnode->table;
				break;
				}
			else if (IsEqual(nodename, "sql"))
				{
				if (oprnode->dbobj->table)
					warningmsg
					("You have specified both a table and a SQL"
					 " statement. This shouldn't be done. Line %d",
					 xmlGetLineNo(node));
				oprnode->dbobj->sql =
				    g_strstrip(xmlNodeGetContent(node));
				oprnode->dbobj->table =
				    db_sqlparse_getfirsttable(oprnode->dbobj->sql);
				oprnode->variablestofree =
				    g_list_append(oprnode->variablestofree,
				                  oprnode->dbobj->table);
				if (!oprnode->table)
					oprnode->table = oprnode->dbobj->table;
				if (!oprnode->primary_table)
					oprnode->primary_table = oprnode->table;
				break;
				}
			else if (IsEqual(nodename, "loadby"))
				{
				oprnode->dbobj->loadby =
				    g_list_append(oprnode->dbobj->loadby,
				                  xmlNodeGetContent(node));
				break;
				}
			else if (IsEqual(nodename, "cache"))
				{
				oprnode->dbobj->cache = TRUE;
				break;
				}
			else if (IsEqual(nodename, "loadall"))
				{
				oprnode->dbobj->loadall = TRUE;
				break;
				}
			else if (IsEqual(nodename, "filter"))
				{
				bondxml_parsewidget(bond, node, ParseFilter, widgetname,
				                    oprnode, parentnode, NULL);
				break;
				}
			else if (IsEqual(nodename, "search"))
				{
				bondxml_parsewidget(bond, node, ParseSearch, widgetname,
				                    oprnode, parentnode, NULL);
				break;
				}
			else if (IsEqual(nodename, "dontexec"))
				{
				oprnode->dbobj->dontexec = TRUE;
				}
			break;
		case ParseLookup:
			if (IsEqual(nodename, "dbobjectsrc"))
				{
				oprnode->dbobjectsrc = xmlNodeGetContent(node);
				break;
				}
			else if (IsEqual(nodename, "dbobject"))
				{
				bondxml_parsewidget_adddb(node, oprnode);
				bondxml_parsewidget(bond, node, ParseDbObject,
				                    widgetname, oprnode, 
						    parentnode, NULL);
				break;
				}
			else if (IsEqual(nodename, "primary"))
				{
				oprnode->primary_table = xmlGetProp(node, "table");
				oprnode->primary = xmlNodeGetContent(node);
				if (!oprnode->primary_table)
					oprnode->primary_table = oprnode->table;
				break;
				}
			else if (IsEqual(nodename, "nulllistvalue"))
				{
				oprnode->nulllistvalue = xmlNodeGetContent(node);
				break;
				}
			else if (IsEqual(nodename, "field"))
				{
				if (oprnode->field)
					multifield = TRUE;
				tmptable = xmlGetProp(node, "table");
				oprnode->field = xmlNodeGetContent(node);
				if (multifield == TRUE)
					{
					if (!tmptable)
						tmptable = oprnode->table;
					content = xmlGetProp(node, "appendtext");
					oprnode->fieldlist =
					    g_list_append(oprnode->fieldlist,
					                  oprnode->field);
					oprnode->tablelist =
					    g_list_append(oprnode->tablelist, 
							    tmptable);
					if (content)
						oprnode->appendtextlist = 
						g_list_append(oprnode->appendtextlist, content);
					}
				break;
				}
			break;
		case ParseFilter:
			if (IsEqual(nodename, "field"))
				{
				tmpfilter->value_field = xmlNodeGetContent(node);
				tmptable = xmlGetProp(node, "table");
				if (tmptable)
					if (strlen(tmptable) > 0)
						tmpfilter->value_table = tmptable;
				break;
				}
			else if (IsEqual(nodename, "widget"))
				{
				tmpfilter->value_widget = xmlNodeGetContent(node);
				break;
				}
			else if (IsEqual(nodename, "constant"))
				{
				tmpfilter->value_constant = xmlNodeGetContent(node);
				break;
				}
			else if (IsEqual(nodename, "dbobjectsrc"))
				{
				tmpfilter->value_dbobjectsrc = xmlNodeGetContent(node);
				tmpnode =
				    bondxml_getnodebyobjectsrc(bond,
				                               tmpfilter->
				                               value_dbobjectsrc);
				if (tmpnode)
					{
					oprnode->dbobjectsrc_widget = 
						tmpnode->widgetname;
					if (!tmpfilter->value_table && 
							tmpnode->dbobj)
						tmpfilter->value_table = 
							tmpnode->dbobj->table;
					}
				else
					errormsg("Unable to find dbobjectsrc "
						"%s in filter", tmpfilter->value_dbobjectsrc);
				break;
				}
			break;
		case ParseSearch:
			if (IsEqual(nodename, "widget"))
				{
				tmpsearch->widgetname = xmlNodeGetContent(node);
				tmpsearch->widget =
				    ui_getwidget(bond->uixml, tmpsearch->widgetname);
				}
			break;
		case ParseTrigger:
			tmptrigger->widgetname = widgetname;
			if (IsEqual(nodename, "function"))
				{
				tmptrigger->functionlanguage =
				    xmlGetProp(node, "language");
				tmptrigger->functionname =
				    xmlGetProp(node, "functionname");
				bondxml_parsewidget(bond, node, ParseFunction,
				                    widgetname, oprnode, parentnode,
				                    tmptrigger);
				}
			else if (IsEqual(nodename, "value"))
				{
				tmptrigger->valuelink = TRUE;
				bondxml_parsewidget(bond, node, ParseFunction,
				                    widgetname, oprnode, parentnode,
				                    tmptrigger);
				}
			else if (IsEqual(nodename, "api"))
				{
				bondxml_parsewidget(bond, node, ParseAPI, 
					widgetname, oprnode, parentnode, tmptrigger);
				}
			else if (IsEqual(nodename, "shell"))
				{
				tmptrigger->shell = xmlGetProp(node, "script");
 				bondxml_parsewidget(bond, node, ParseFunction,
				                    widgetname, oprnode, parentnode,
				                    tmptrigger); 
				}
			else if (IsEqual(nodename, "papyrus"))
				{
				tmptrigger->runpapyrus = TRUE;
				}
			else if (IsEqual(nodename, "report"))
				{
				if (xmlGetProp(node, "file"))
					tmptrigger->reportfile = xmlGetProp(node, "file");
				/* Tim: another hack I'm not proud of */
				else
					tmptrigger->reportfile = "";

				tmptrigger->reportfiletype = xmlGetProp(node, "type");
				content = xmlGetProp(node, "preview");
				if (content && IsTrue(content))
					tmptrigger->reportpreview = TRUE;
				bondxml_parsewidget(bond, node, ParseFunction,
				                    widgetname, oprnode, parentnode,
				                    tmptrigger);

				bondxml_parsewidget(bond, node, ParseReportfile,
				                    widgetname, oprnode, parentnode,
				                    tmptrigger);

				}
			break;
		case ParseReportfile:
			if (IsEqual(nodename, "reportfile"))
				{
				bondxml_parsewidget(bond, node, ParseArgument,
				                    widgetname, oprnode, parentnode,
				                    trigger);
				}
			break;
		case ParseFunction:
			if (IsEqual(nodename, "argument"))
				{
				bondxml_parsewidget(bond, node, ParseArgument,
				                    widgetname, oprnode, parentnode,
				                    trigger);
				}
			break;
		case ParseArgument:
			if (IsEqual(nodename, "widget"))
				{
				/* If this is within reportfile tag, not an 
				 * arg but rather the reportfile name */
				if (IsEqual(parentnode->name, "reportfile"))
					{
					tmptrigger->reportfile = xmlNodeGetContent(node);
					debuglog(100, "report file name: %s",
					         tmptrigger->reportfile);
					}
				else
					tmpargument->widgetname = xmlNodeGetContent(node);
				}
			else if (IsEqual(nodename, "constant"))
				{
				tmpargument->constant = xmlNodeGetContent(node);
				}
			else if (IsEqual(nodename, "dbobjectsrc"))
				{
				tmpargument->dbobjectsrc = xmlNodeGetContent(node);
				if (!oprnode->dbobjectsrc)
					oprnode->dbobjectsrc = tmpargument->dbobjectsrc;
				}
			else if (IsEqual(nodename, "field"))
				{
				tmpargument->field = xmlNodeGetContent(node);
				tmptable = xmlGetProp(node, "table");
				if (!tmptable)
					tmptable = oprnode->table;
				tmpargument->table = tmptable;
				}
			break;
		case ParseAPI:
			if (IsEqual(nodename, "widget"))
				{
				trigger->apiwidgetname = xmlNodeGetContent(node);
				/* debuglog(50, "api call on widget %s",
				         trigger->apiwidgetname); */
				}
			else if (IsEqual(nodename, "loadby"))
				{
				trigger->loadby =
				    g_list_prepend(trigger->loadby,
				                   xmlNodeGetContent(node));
				}
			else if (IsEqual(nodename, "sql"))
				{
				trigger->sql = xmlNodeGetContent(node);
				}
			break;
		default:
			warningmsg("Unknown parse state, stuck at line %d",
			           xmlGetLineNo(node));
			break;
			}
		}

	if (state == ParseArgument)
		{
		bondxml_parsewidget_findfieldpath(oprnode, tmpargument);
		}

	if (addtohash == TRUE)
		{
		mem_verify(oprnode);
		bondxml_addfieldinfo(bond, oprnode);
		/* hash entries are done from the bottom up to save doing a double
		   parse */
		g_hash_table_insert(bond->hashoprnode, widgetname, oprnode);
		if (oprnode->dbobj)
			g_hash_table_insert(bond->hashdbobjectsrc,
			                    oprnode->dbobjectsrc, oprnode);
		}

	return oprnode;
	}

/* get dbobjectsrc_widget for every widget in xml file */
static void
bondxml_secondparse(Bond * bond, OperationsNode * oprnode,
                    OperationsNode * lastnode, gchar * lastdbobjectsrc)
	{
	OperationsNodeTriggerArgument *argument;
	OperationsNodeTrigger *trigger;
	OperationsNodeFilter *filternode;
	OperationsNode *tmpnode;
	GList *walk, *inwalk;

	mem_verify(oprnode);
	/* find the widgetname for dbobjectsrc. */
	if (oprnode->dbobjectsrc)
		{
		if (lastdbobjectsrc && IsEqual(oprnode->dbobjectsrc, lastdbobjectsrc)
		        && lastnode)
			oprnode->dbobjectsrc_widget = lastnode->widgetname;

		else
			{
			tmpnode = bondxml_getnodebyobjectsrc(bond, oprnode->dbobjectsrc);
			if (!tmpnode)
				{
				errormsg("No dbobjectsrc found for %s",
				         oprnode->widgetname);
				return ;
				}
			oprnode->dbobjectsrc_widget = tmpnode->widgetname;
			lastdbobjectsrc = tmpnode->widgetname;
			lastnode = tmpnode;
			}
		}
	/* Parse the arguments in triggers again to make sure we have all the
	   details */
	if (oprnode->triggerlist)
		for (walk = g_list_first(oprnode->triggerlist); walk; walk = walk->next)
			{
			trigger = walk->data;
			for (inwalk = g_list_first(trigger->argumentlist); inwalk;
			        inwalk = inwalk->next)
				{
				argument = inwalk->data;
				if (argument->fieldpathtomake == FALSE)
					continue;
				if (argument->table && argument->field && oprnode->field &&
				        oprnode->table)
					{
					argument->fieldpathlist =
					    fieldpath_init(oprnode->table, oprnode->field,
					                   argument->table);
					argument->fieldpathtomake = FALSE;
					continue;
					}
				if (argument->dbobjectsrc)
					tmpnode =
					    bondxml_getnodebyobjectsrc(bond, oprnode->dbobjectsrc);
				else if (oprnode->dbobj)
					tmpnode = oprnode;
				else if (oprnode->dbobjectsrc)
					tmpnode =
					    bondxml_getnodebyobjectsrc(bond, oprnode->dbobjectsrc);
				else
					tmpnode = NULL;
				if (tmpnode)
					{
					if (tmpnode->table)
						{
						argument->fieldpathlist =
						    fieldpath_init(tmpnode->table, tmpnode->field,
						                   argument->table);
						argument->fieldpathtomake = FALSE;
						continue;
						}
					warningmsg
					("Please specify a table on database source %s"
					 " because a trigger can't work out its orgin.",
					 tmpnode->widgetname);
					}
				}
			}
	/* if filter on object find the objectsrc_widget for it */
	if (oprnode->dbobj && oprnode->dbobj->filters)
		for (walk = g_list_first(oprnode->dbobj->filters); walk;
		        walk = walk->next)
			{
			filternode = walk->data;
			if (filternode->value_dbobjectsrc)
				{
				tmpnode =
				    bondxml_getnodebyobjectsrc(bond, oprnode->dbobjectsrc);
				filternode->value_dbobjectsrc_widget = tmpnode->widgetname;
				}
			}

	for (walk = g_list_first(oprnode->childrennode); walk; walk = walk->next)
		bondxml_secondparse(bond, walk->data, lastnode, lastdbobjectsrc);
	}

/**
 * bondxml_parse:
 * @bond: Bond structure
 * @node: Node in the xml parsing tree
 * @oprparent: Current @node's operations node parent
 *
 * Recursive function for parsing widgets.
 *
 * Returns: 0
 */
gint
bondxml_parse(Bond * bond, xmlNode * node, OperationsNode * oprparent)
	{
	gchar *longname;
	OperationsNode *oprprev = NULL, *oprnode;
	GList *tmplist = NULL, *walk;
	
	/* For each child of @node */
	for (node = node->children; node; node = node->next)
		{
		oprnode = NULL;
		longname = xmlGetProp(node, "longname");
		/* Collect information on all widget nodes that have longname
		 * attached to them. */
		if (longname)
			{
			/* Parse the nodes */
			oprnode = bondxml_parsewidget(bond, node, ParseStart, 
					longname, NULL, oprparent, NULL);
			if (!oprnode)
				warningmsg("Unable to create an operations "
					"node on %s at line %d", longname, 
					xmlGetLineNo(node));
			else
				{
				/* Add the newly created node into tree. */
				oprnode->prevnode = oprprev;
				if (oprprev)
					oprprev->nextnode = oprnode;
				oprprev = oprnode;
				}
			}
		/* Recursively parse the children of this operations node 
		 * only if a widget that has longname attached has been found. 
		 * Otherwise continue parsing with current node's operational 
		 * parent (an input argument to this function). */
		if (oprnode)
			bondxml_parse(bond, node, oprnode);
		else
			bondxml_parse(bond, node, oprparent);
		/* If at top of the tree remember the parents. As you will 
		 * double parse them again. */
		if (!oprparent && oprnode)
			tmplist = g_list_prepend(tmplist, oprnode);
		}
	if (tmplist)
		{
		for (walk = tmplist; walk; walk = walk->next)
			bondxml_secondparse(bond, walk->data, NULL, NULL);
		g_list_free(tmplist);
		}

	return 0;
	}

/**
 * bondxml_getwidgetlongname:
 * @bond: Bond xml document.
 * @widgetshortname: Widget name to look up.
 * 
 * Finds the long name of a widget if it exist, else if it a widget long
 * name is actually passed in it will return that instead. 
 * The return value is not to be freed.
 * 
 * Returns: %NULL if @widgetshortname is not a valid widget else the name of the widget.
 */
gchar *
bondxml_getwidgetlongname(Bond * bond, gchar * widgetshortname)
	{
	UiWidget *widget;

	g_assert(widgetshortname);
	g_assert(bond);

	widget = g_hash_table_lookup(bond->uixml->longname, widgetshortname);
	/* Check if shortname was found in hashtable. If this fails then 
	 * load then look for widget by long name. */
	if (widget)
		return g_hash_table_lookup(bond->uixml->widgets, widget);
	else
		widget = g_hash_table_lookup(bond->uixml->shortname, widgetshortname);

	if (widget)
		return g_hash_table_lookup(bond->uixml->widgets, widget);
	warningmsg("Failed to find widget %s", widgetshortname);
	return NULL;
	}

static void
bondxml_debug_rec(OperationsNode * node, gint indent)
	{
	GList *walk;
	OperationsNodeFilter *filter;

	if (!node)
		{
		debug_output("%*sNULL oprnode. Ouch.\n", indent, "");
		return ;
		}
	debug_output("%*s%s\n", indent, "", node->widgetname);
	if (node->field)
		debug_output("%*sField: %s from %s\n", indent, "", node->field,
		             node->dbobjectsrc);
	if (node->primary)
		debug_output("%*sPrimary: %s\n", indent, "", node->primary);
	if (node->nullvalue)
		debug_output("%*sNullvalue: %s\n", indent, "", node->nullvalue);
	if (node->parent)
		debug_output("%*sParent: %s\n", indent, "", (gchar *) 
				node->parent->data);
	if (node->dbobj)
		{
		debug_output("%*sdbobject: %s from %s\n", indent, "",
		             node->dbobj->name, node->dbobj->table);
		for (walk = g_list_first(node->dbobj->filters); walk;
		        walk = walk->next)
			{
			filter = walk->data;
			g_assert(filter);
			debug_output("%*sFilter: field: %s", indent, "", 
					filter->field);
			if (filter->value_constant)
				debug_output("=%s\n", filter->value_constant);
			else if (filter->value_widget)
				debug_output("=%s\n", filter->value_widget);
			else if (filter->value_field)
				debug_output("=%s\n", filter->value_field);
			else
				debug_output("=Unknown?\n");
			}
		}

	for (walk = g_list_first(node->childrennode); walk; walk = walk->next)
		bondxml_debug_rec(walk->data, indent + 2);
	}

void
bondxml_debug(OperationsNode * node)
	{
	debug_minorheader();
	bondxml_debug_rec(node, 0);
	debug_minorfooter();
	}

gint
bondxml_test(Bond * bond, gint testnum)
	{
	gint retval = 0, passed = 0;
	OperationsNode *oprnode;

	oprnode = bondxml_getnode(bond, "mainwindow");
	g_assert(oprnode);
	switch (testnum)
		{
	case 0:
		test_annonce(0, "OperationsNode parsing");
		if (oprnode->dbobj)
			passed++;
		if (oprnode->widgettype == UiForm)
			passed++;
		else
			errormsg("Widget %s isn't a window", oprnode->widgetname);
		test_result(oprnode->widgetname, "mainwindow", &passed);
		test_result(oprnode->dbobj->name, "soldiers", &passed);
		test_result(oprnode->dbobj->table, "soldier", &passed);
		retval = test_total(passed, 5);
		break;
	case 1:
		test_annonce(1, "OperationsNode parsing");
		oprnode = bondxml_getnode(bond, "mainwindow");
		bondxml_debug(oprnode);
		break;
		}
	return retval;
	}
