/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "bc.h"
#include "db.h"

#include "fieldpath.h"

/**
 * This code has been ported from bond 0.4.4, which in it was called iMethodBuild.
 * I've included this documentation because its still relevant but it uses the old
 * naming conventions. Just replace imethodbuild with fieldpath. Aliases and imethod
 * no longer exist. Aliases was replaced with bondobjects and imethod by triggers.
 * 
 * iMethod - UiGeneral documentation
 * imethodbuild.c provides the components to build information used by imethod.c,
 * imethod.c is the more runtime of the components dealing of what happens when
 * the user clicks somewhere while imethodbuild deals with the initial information
 * gathering.
 *
 * Aliases (AliasDataSource*) may point to functions, if an alias points to 
 * a function then detailed information needs to be stored about that function 
 * which it is pointing to.  A function is a user defined C function method 
 * that is to be called at runtime under certain circumstances.  A gchar* or
 * a database object is returned by the function depending on what type of
 * widget it is on. 
 *
 * Widgetnames to detonate a function would look like this
 * funcname(tableb.fieldname,tableb.otherfield,awidgetname);
 *
 * Functions are called with arguments, which can be widget names or table.fieldnames.
 * Every arguments that has a corresponding widget needs information stored in its
 * own alias (AliasDataSource*).  If a widget controls the values that appear in
 * other widgets, ie is a linkwidget, then it also needs information stored about it.
 * Ie the value selected in a certain widget controls what the final argument
 * is passed onto a function so when that value is changed it needs to know
 * in the alias it needs to re-run any children functions.
 * 
 * This leads to 3 principle types of iMethod structures in AliasDataSources.
 * Child: The alias for the widget that is actually the function.
 * Links: Functions that link the form where the function appears to its arguments. Often
 * these are dropdown boxes that represents a constraint relationship.
 * Masters: These are the final widgets/table.fields which are the desired arguments
 * to pass onto the child function.
 *
 * iMethod is designed to cope with changing data yet to be written to the backend
 * database.  The links play an important role in this when dealing with UI changes,
 * multiple forms need to be brought together to manage changes correctly.  
 * Each iMethod stores both its table.field description, of where it extracts data
 * from and information about what widget also contains this information.
 * 
 * When it comes to imethod_build_all(), everytime you display a new widget, and
 * find the alias with it, using the alias_finddatasources() function, it needs
 * to update records that could be effected by it.  For example form X wasn't
 * visible or previously opened before, but now it is. This table is referenced
 * by a method call else where in the bond app, the method code needs to refresh its
 * knowledge of all the methods.  So it needs to know that this method call on this
 * other form can now get its information from a new widget just displayed as well
 * as from the database.
 * 
 * ps. this module was not in anyway influnced by iMac's.
 *
 * --------------------------------------------------------
 * Francis: I would like to comment on a few things.
 *
 * Disclaimer:
 * This is much based on my understanding and guessing consult with Andrew if you
 * want to know really deeply.
 *
 * Comment and Documentation:
 *   The description of functions written by Francis has a tag 'Francis:'.
 *
 * Terminology:
 *   method-widget; a bit special widget which works as method invocation.
 *   
 * Naming convention:
 *   1. Functions with suffix '_rec' indicate they are recursive.
 *   2. Static functions like imethod_init_imethod_create () is a sub-function of
 *      the imethod_init() function.
 *
 * External interfaces:
 *   The most important functions are
 *   1. imethod_init()
 *   2. imethod_build_all()
 *
 *   imethod_init():
 *
 */


static gboolean
fieldpath_createallpath_alreadystored(FieldPath * path, DbConstraintDef * c)
	{
	GList *walk;

	g_assert(path);
	g_assert(c);
	if (path->parent == NULL)
		return FALSE;
	for (walk = g_list_first(path->parent->allcrefsdone); walk != NULL; 
		walk = walk->next)
		if (walk->data == c)
			return TRUE;
	return FALSE;
	}

static FieldPath *
fieldpath_createpath_element(gchar * fieldname, gchar * tablename)
	{
	FieldPath *path;
	path = (FieldPath *) mem_calloc(sizeof(FieldPath));
	path->fieldname = fieldname;
	path->tablename = tablename;
	path->parent = path;
	return path;
	}

static void
fieldpath_freepath_elements_rec(FieldPath *path, gboolean inusetofree)
	{
	GList *walk;
	for(walk = g_list_first(path->next);walk!=NULL;walk=walk->next)
		fieldpath_freepath_elements_rec(walk->data, inusetofree);
	/* m.essage("freeing memory id %d",mem_getid(path)); */
	if (inusetofree == FALSE && path->inuse == TRUE)
		;
	else
		mem_free(path);
	}
	
static gint
fieldpath_freepath_elements(FieldPath *completepath, gboolean inusetofree)
	{
	if (!completepath)
		return -1;
	if (completepath->parent != completepath)
		return -2;
	
	fieldpath_freepath_elements_rec(completepath,inusetofree);
	return 0;
	}

/* Build a complete list of what tables refer to what. Its a total map of
 * all relationships. */
static void
fieldpath_createallpath_rec(FieldPath * path, GList * crefmaster, gchar *fieldstr, gchar * tablename)
	{
	DbConstraintDef *c;
	GList *walk, *creflist;
	FieldPath *childpath;

	if (crefmaster == NULL)
		crefmaster = db_constraint_getall();
	message("Path %s.%s", path->tablename, path->fieldname);
	/* find all relationships, add them */
	creflist = db_constraint_getontable(crefmaster, tablename);
	for (walk = g_list_first(creflist); walk != NULL; walk = walk->next)
		{
		c = walk->data;
		if (path->prev != NULL)
			if (strcmp(c->column[1], path->prev->fieldname) == 0 && 
				strcmp(c->table[1], path->prev->tablename) == 0)
				continue;
		if (c->casual == 1)
			continue;
		if (fieldpath_createallpath_alreadystored(path, c) == TRUE)
			continue;
		message("Adding %s.%s -> %s.%s", c->table[0], c->column[0], 
			c->table[1], c->column[1]);
		/* Create the child which we shall branch into and generate path details 
		 * in a recusive fashion 
		 */
		childpath = mem_calloc(sizeof(FieldPath));
		childpath->parent = path->parent;
		childpath->prev = path;
		childpath->c = c;
		childpath->fieldname = c->column[1];
		childpath->tablename = c->table[1];
		path->parent->allcrefsdone = g_list_prepend(path->parent->allcrefsdone, c);
		path->next = g_list_prepend(path->next, childpath);
		/* Call myself. cause i am just so damn sexy */
		/* message("prev stuff %s.%s", childpath->prev->tablename, childpath->prev->fieldname); */
		fieldpath_createallpath_rec(childpath, crefmaster, c->column[1], c->table[1]);
		}
	g_list_free(creflist);
	}

/**
 * find shortest path recusive function, the arr amd num are the finial results.
 * Ohhh this code looks sexy. esp the (*arr)[(*num)-1] bit.
 */
static gint
fieldpath_findshortestpath_rec(FieldPath * path, gchar * desttable, GList *** arr, gint * num, gboolean start)
	{
	FieldPath *current;
	gint retval = 0, cur_retval = 0;
	GList *walk;

	debuglog(40,"Finding shortest path to %s, currently at %s", desttable, 
		path->tablename);
	/* are we there yet? */
	if (strcmp(desttable, path->tablename) == 0)
		{
		(*num)++;
		(*arr) = (GList **) mem_realloc((*arr), sizeof(GList **) * ((*num) + 1));
		(*arr)[((*num) - 1)] = NULL;		 /* i am working on this code see */
		(*arr)[((*num) - 1)] = g_list_prepend((*arr)[((*num) - 1)], path);
		debugmsg("Returning. Found match on %s=%s", desttable, path->tablename);
		return 1;
		}
	if ((*arr) == NULL)
		{
		(*arr) = (GList **) mem_calloc(sizeof(GList **));
		*num = 0;
		debugmsg("allocating memory now and setting num to %d. "
			"Looking for %s from %s", *num, desttable,path->tablename);
		}
	for (walk = g_list_first(path->next); walk != NULL; walk = walk->next)
		{
		current = walk->data;
		debugmsg("Constraints: %s.%s->%s.%s", current->c->table[0], 
			current->c->column[0], current->c->table[1],current->c->column[1]);
		if ((cur_retval = fieldpath_findshortestpath_rec(
				current, desttable, arr, num, FALSE)) >= 1)
			{
			cur_retval++;
			if (cur_retval > retval)
				retval = cur_retval;
			/* Add to or create a glist of paths that are used to reach this 
			 * destination in order of how they got there 
			 */
			(*arr)[((*num) - 1)] = g_list_prepend((*arr)[((*num) - 1)], path);
			if (cur_retval != (g_list_length((*arr)[((*num) - 1)])))
				warningmsg("This will be dropped cause it doesn't match "
					"the desired finial path .%s.%s",current->c->table[0], 
					current->c->column[0]);
			debugmsg("Num is at %d, retval at %d, %d length", (*num), 
				retval, g_list_length((*arr)[((*num) - 1)]));
			}
		}
	debugmsg("At pass %s . Num :%d, Retval :%d", path->tablename, (*num), retval);
	return retval;
	}

static void
fieldpath_debugpath(FieldPath * path, gint indent)
	{
	GList *walk;
	FieldPath *cur = NULL, *old = NULL;

	g_assert(path);
	for (walk = g_list_first(path->next); walk != NULL; walk = walk->next)
		{
		cur = walk->data;
		old = cur->prev;
		if (old != NULL)
			debuglog(40,"%*s%s.%s from %s.%s", indent * 2, "", cur->tablename, 
				cur->fieldname, old->tablename,old->fieldname);
		else
			debuglog(40,"%*s%s.%s", indent * 2, "", cur->tablename, cur->fieldname);
		fieldpath_debugpath(cur, indent + 1);
		}
	}

static void
fieldpath_findshortestpath_debug(GList * pathlist)
	{
	GList *walk;
	FieldPath *path;

	debugmsg("%d items in total", g_list_length(pathlist));
	for (walk = g_list_first(pathlist); walk != NULL; walk = walk->next)
		{
		path = walk->data;
		if (path->c != NULL)
			debuglog(40,"%s.%s get from %s.%s", path->tablename, path->fieldname, 
				path->c->table[0],path->c->column[0]);
		else
			debuglog(40,"%s.%s is starting table,", path->tablename, path->fieldname);
		}
	}

/**
 * fieldpath_debug:
 * @pathtotable:
 * 
 * Show the path orgins.
 */
void 
fieldpath_debug(GList *pathtotable)
	{
	fieldpath_findshortestpath_debug(pathtotable);
	}
	
static FieldPath *
fieldpath_findpathfromsource(FieldPath * completepath, gchar * sourcetable)
	{
	if (strcmp(completepath->tablename, sourcetable) == 0)
		return completepath;
	/* todo, a more detailed searching method */
	warningmsg("The complete path isn't pointing to its origins. %s != %s",
		completepath->tablename, sourcetable);
	return NULL;
	}

/* Find the shortest path between @sourcetable and @desttable and return a GList of
 * iMethodPath* to describe that path.  Return the glist returned when your done with it.
 */
static GList *
fieldpath_findshortestpath(FieldPath * completepath, gchar * sourcetable, gchar * desttable)
	{
	GList **arr = NULL, *retlist = NULL;
	GList *walk;
	gint num = 0, i, max = 65536 / 2;
	FieldPath *path;

	path = fieldpath_findpathfromsource(completepath, sourcetable);
	if (path == NULL)
		return NULL;
	if (fieldpath_findshortestpath_rec(path, desttable, &arr, &num, TRUE) == 0)
		{
		if (arr != NULL)
			mem_free(arr);
		warningmsg("No path was found from %s to %s", sourcetable, desttable);
		return NULL;
		}
	/* Find the actual shortest path by measuring the length of the list */
	debugmsg("\n\nFound %d paths to %s from %s", num, desttable, sourcetable);
	if (arr != NULL)
		for (i = 0; i < num; i++)
			{
			if (arr[i] == NULL)
				continue;
			fieldpath_findshortestpath_debug(arr[i]);
			if (g_list_length(arr[i]) < max)
				{
				max = g_list_length(arr[i]);
				if (retlist != NULL)
					g_list_free(retlist);
				retlist = arr[i];
				}
			else
				g_list_free(arr[i]);
			}
	else
		warningmsg("arr was NULL but this shouldn't be NULL. "
			"This was in finding %s", desttable);
	/* Mark whats going to be returned so we can free the rest. Unwanted paths
	 * will be freed later on.	*/
	if (retlist)
		for (walk=g_list_first(retlist);walk!=NULL;walk=walk->next)
			{
			path = walk->data;
			path->inuse = TRUE;
			}
	/* Free the array and the not needed items */
	if (arr != NULL)
		{
		for (i=0;i<num;i++)
			{
			/* Only proceed if we are not the return list and we are a valid list */
			if (arr[i] != retlist && arr[i])
				{
				/* free the actual list */
				g_list_free(arr[i]);
				}			
			}
		mem_free(arr);
		}
	/* Return the shortest path else %NULL on failure */
	return retlist;
	}

/**
 * fieldpath_init:
 * @tablesrc:
 * @fieldsrc: 
 * @tabledest:
 *
 * Finds the path from tablesrc.fieldsrc to tabledest across relationships
 * defined in the database.
 *
 */
GList *
fieldpath_init(gchar *tablesrc, gchar *fieldsrc, gchar *tabledest)
	{
	FieldPath *path;
	GList *pathtotable;
	/* Create the start of the path */
	path = fieldpath_createpath_element(fieldsrc, tablesrc);
	/* Find all possible path options */
	fieldpath_createallpath_rec(path, NULL, fieldsrc, tablesrc);
	/* Find the shortest distance of all the paths */
	pathtotable = fieldpath_findshortestpath(path, tablesrc, tabledest);
	/* Free all elements execpt those that we are returning */
	fieldpath_freepath_elements(path, FALSE);

	return pathtotable;
	}
	
/**
 * fieldpath_free:
 * @pathtotable: GList* to free.
 *
 * Frees contents created by fieldpath_init().
 *
 */
void
fieldpath_free(GList *pathtotable)
	{
	FieldPath *path;
	GList *walk;
	for (walk=g_list_first(pathtotable);walk!=NULL;walk=walk->next)
		{
		path = walk->data;
		mem_free(path);
		}
	g_list_free(pathtotable);
	}
	
gint
fieldpath_test(gint testnum)
	{
	gchar *fieldsrc, *tablesrc, *fielddest, *tabledest;
	FieldPath *path;
	GList *pathtotable;
	
	fieldsrc = mem_strdup("kills");
	tablesrc = mem_strdup("soldier");
	fielddest = mem_strdup("pay");
	tabledest = mem_strdup("rank");
	path = fieldpath_createpath_element(fieldsrc, tablesrc);
	fieldpath_createallpath_rec(path, NULL, fieldsrc, tablesrc);
	debugmsg("showing all paths");
	fieldpath_debugpath(path, 0);
	debugmsg("finding shortest path now");
	pathtotable = fieldpath_findshortestpath(path, tablesrc, tabledest);
	debugmsg("showing debug contents");
	fieldpath_findshortestpath_debug(pathtotable);
	
	mem_free(fielddest);
	mem_free(tabledest);
	mem_free(fieldsrc);
	mem_free(tablesrc);
	return 0;
	}
