/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <stdio.h>
#include <glob.h>
#include <dlfcn.h>
#include <string.h>

#include "ui.h"
#include "uiplugin.h"
#include "dropdown.h"

#define NUMGLADESEARCH 4

gchar *gladefilesearch[NUMGLADESEARCH] = {
            "./",
            "/usr/local/share/",
            "/usr/share/",
            "../share/"
        };

/**
 * ui_init:
 * @library:
 * @args:
 *
 * Initialises user interface library, loading the library up and execting
 * any of its user interface initisalation code that needs running.
 *
 * Returns: %NULL on error.
 */
UiXML *
ui_init(gchar * library, gchar * localargs, gchar * xmlfile, gint argc,
        gchar * argv[])
	{
	gint i;
	UiXML *uixml = NULL;
	Plugin *plugin;
	PluginFunction *f = NULL;
	typedef UiXML *(*function) (gchar *, gint, gchar **);
	function init;

	/* Load the plugin */
	plugin = plugin_create(library, localargs);
	if (!plugin)
		return NULL;

	for (i = 0; i < plugin->numfunctions; i++)
		if (strcmp("init", plugin->func[i]->name) == 0)
			{
			f = plugin->func[i];
			break;
			}
	g_assert(f);
	/* Run the in it in the plugin library. */
	init = f->ptr;
	xmlfile = ui_findxmlfile(xmlfile);
	if (xmlfile)
		uixml = init(xmlfile, argc, argv);
	else
		{
		plugin_free(plugin);
		return NULL;
		}
	if (!uixml)
		{
		plugin_free(plugin);
		return NULL;
		}

	uixml->plugin = plugin;
	return uixml;
	}

/**
 * ui_cleanup:
 * 
 * Frees up and closes down the user interface. 
 */
gint
ui_cleanup(UiXML * uixml)
	{
	gint retval;

	mem_freegarbage(uixml->garbage);
	retval = plugin_exec(uixml, "cleanup");
	plugin_free(uixml->plugin);
	uixml->plugin = NULL;
	mem_free(uixml);
	return retval;
	}

/**
 * ui_main:
 *
 * Enter the main loop.
 *
 */
gint
ui_main(UiXML * uixml)
	{
	return plugin_exec(uixml, "main");
	}

/**
 * ui_mainquit:
 *
 * Quit the main loop.
 *
 */
gint
ui_mainquit(UiXML * uixml)
	{
	return plugin_exec(uixml, "mainquit");
	}

/**
 * ui_getparentformname:
 * @widget: A widget.
 *
 * Find the parent widget, and get its name.
 */
gchar *
ui_getparentformname(UiXML * uixml, UiWidget * widget)
	{
	return plugin_exec_arg1(uixml, "getparentformname", widget);
	}

UiWidget *
ui_getwidget(UiXML * uixml, gchar * widgetname)
	{
	static PluginFunction *f;
	typedef UiWidget *(*function) (UiXML * uixml, gchar * widgetname);
	function func;

	if (!f)
		if ((f = plugin_findfunction(uixml, "getwidget")) == NULL)
			return NULL;
	func = f->ptr;
	return func(uixml, widgetname);
	}

gchar *
ui_getname(UiXML * uixml, UiWidget * widget)
	{
	return plugin_exec_arg1(uixml, "getname", widget);
	}

gchar *
ui_getname_short(UiXML * uixml, UiWidget * widget)
	{
	return plugin_exec_arg1(uixml, "getname_short", widget);
	}

void
ui_parse(UiXML * uixml)
	{
	plugin_exec(uixml, "parse");
	}

gint
ui_getvalue(UiXML * uixml, UiWidget * widget, gchar ** value)
	{
	static PluginFunction *f;
	typedef gint(*function) (UiXML * uixml, UiWidget * widget, gchar **);
	function func;

	if (!f)
		if ((f = plugin_findfunction(uixml, "getvalue")) == NULL)
			return -1;
	func = f->ptr;
	return func(uixml, widget, value);
	}

gint
ui_getvaluebyname(UiXML * uixml, gchar * widgetname, gchar ** value)
	{
	UiWidget *widget;
	static PluginFunction *f;
	typedef gint(*function) (UiXML * uixml, gchar * widgetname, gchar ** value);
	function func;

	widget = ui_getwidget(uixml, widgetname);
	g_assert(widget);
	if (!f)
		if ((f = plugin_findfunction(uixml, "getvalue")) == NULL)
			return -1;
	func = f->ptr;
	return func(uixml, widget, value);
	}

/**
 * Francis: This abstract interface is very useful simple widgets such
 * as text entry and so on, but for multi-valued widgets, such as
 * dropdown, columned list, this interface is not sufficient.
 */
gint
ui_setvalue(UiXML * uixml, UiWidget * widget, gchar * value)
	{
	static PluginFunction *f;
	typedef gint(*function) (UiXML * uixml, UiWidget * widget, gchar * value);
	function func;

	if (!f)
		if ((f = plugin_findfunction(uixml, "setvalue")) == NULL)
			return -1;
	func = f->ptr;
	return func(uixml, widget, value);
	}

gint
ui_setvaluebyname(UiXML * uixml, gchar * widgetname, gchar * value)
	{
	return (gint) plugin_exec_arg2(uixml, "setvalue",
	                               ui_getwidget(uixml, widgetname), value);
	}

UiWidgetType
ui_gettype(UiXML * uixml, UiWidget * widget)
	{
	static PluginFunction *f;
	typedef UiWidgetType(*function) (UiXML * uixml, UiWidget * widget);
	function func;

	if (!f)
		if ((f = plugin_findfunction(uixml, "gettype")) == NULL)
			return -1;

	func = f->ptr;
	return func(uixml, widget);
	}

/**
 * ui_widgetchildren:
 * @widgetlongname:
 *
 * Get a list of all widgets that are inside of @widgetlongname.
 *
 * Returns: %NULL on error.
 */
GList *
ui_widgetchildren(UiXML * uixml, gchar * widgetlongname)
	{
	return plugin_exec_arg1(uixml, "widgetchildren", widgetlongname);
	}

/**
 * ui_showwidget:
 * @uixml: UiXml 
 * @widget: Widget to be made visible.
 * 
 * map to uigtk_showwidget() etc.
 */
void
ui_showwidget(UiXML * uixml, UiWidget * widget)
	{
	return (void)plugin_exec_arg1(uixml, "showwidget", widget);
	}

/**
 * ui_hidewidget:
 *
 * map to uigtk_hidewidget() etc.
 */
void
ui_hidewidget(UiXML * uixml, UiWidget * widget)
	{
	return (void)plugin_exec_arg1(uixml, "hidewidget", widget);
	}

/**
 * ui_widgetlongname:
 *
 * Found that this was missing...
 */
gchar *
ui_widgetlongname(UiXML * uixml, gchar * widgetname)
	{
	return (gchar *) plugin_exec_arg1(uixml, "widgetlongname", widgetname);
	}

/**
 * ui_issignalattached:
 *
 * This is new version of checkifsignalattached() in old 'globalgtk.c'
 */
gint
ui_issignalattached(UiXML * uixml, UiWidget * widget, gchar * name)
	{
	return (gint) plugin_exec_arg2(uixml, "issignalattached", widget, name);
	}

/**
 * ui_isvisible:
 *
 */
gboolean
ui_isvisible(UiXML * uixml, UiWidget * widget)
	{
	static PluginFunction *f;
	typedef gboolean(*function) (UiXML * uixml, UiWidget * widget);
	function func;

	if (!f)
		if ((f = plugin_findfunction(uixml, "isvisible")) == NULL)
			return -1;

	func = f->ptr;
	return func(uixml, widget);
	}

/**
 * ui_getcontainerchildren:
 *
 */
GList *
ui_getcontainerchildren(UiXML * uixml, UiWidget * widget)
	{
	return plugin_exec_arg1(uixml, "getcontainerchildren", widget);
	}

/**
 * ui_signalconnect:
 * @object:
 * @name:
 * 
 * Creates a signal on a widget @object by the name @name with callbacks
 * to function @argfunc and usedata @func_data.
 * 
 * Returns: Signal ID (use to free signal). 
 */
gint
ui_signalconnect(UiXML * uixml, UiWidget * object, const gchar * name,
                 void *(*argfunc) (), void *func_data)
	{
	static PluginFunction *f;
	typedef guint(*function) (UiXML * uixml, UiWidget * object,
	                          const gchar * name, void *(*argfunc) (),
	                          void *func_data);
	function func;

	if (!f)
		if ((f = plugin_findfunction(uixml, "signalconnect")) == NULL)
			return -1;
	func = f->ptr;

	return func(uixml, object, name, argfunc, func_data);
	}
/**
 * ui_signaldisconnectbyfunc:
 * 
 * This function doesn't appear to work properly, caused by a bug in gtk. Use
 * ui_signaldisconnect() instead.
 */	

void
ui_signaldisconnectbyfunc(UiXML * uixml, UiWidget * object, void (*argfunc) (),
                          void *data)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * object,
	                          void (*argfunc) (), void *data);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "signaldisconnectbyfunc");
	func = f->ptr;
	func(uixml, object, argfunc, data);
	}
	
void
ui_signaldisconnect(UiXML * uixml, UiWidget * object, gint signalid)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * object,gint);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "signaldisconnect");
	func = f->ptr;
	func(uixml, object, signalid);
	}

UiSignalKey *
ui_signalconnect_keypress(UiXML * uixml, UiWidget * widget,
                          gint(*callback) (void *, gchar * str), void *userdata)
	{
	UiSignalKey *uisignalkey;
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * widget,
	                          UiSignalKey * uisignalkey);
	function func;

	uisignalkey = mem_calloc(sizeof(UiSignalKey));
	uisignalkey->uixml = uixml;
	uisignalkey->userdata = userdata;
	uisignalkey->func = callback;
	uisignalkey->widget = widget;
	uixml->uisignalkeylist = g_list_prepend(uixml->uisignalkeylist, uisignalkey);

	if (!f)
		f = plugin_findfunction(uixml, "signalconnect_keypress");
	func = f->ptr;

	func(uixml, widget, uisignalkey);
	return uisignalkey;
	}

void
ui_signaldisconnect_keypress(UiXML * uixml, UiWidget * object,
                             UiSignalKey * uisignalkey)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * object,
	                          UiSignalKey * uisignalkey);
	function func;

	/* Return if function not already connected */
	if (!uisignalkey)
		return ;
	if (!f)
		f = plugin_findfunction(uixml, "signaldisconnect_keypress");
	func = f->ptr;
	func(uixml, object, uisignalkey);
	mem_free(uisignalkey);
	}

gint
ui_signalconnect_keypress_free(UiXML * uixml, UiWidget * widget)
	{
	GList *walk;
	UiSignalKey *uisignalkey;

	for (walk = g_list_first(uixml->uisignalkeylist); walk != NULL;
	        walk = walk->next)
		{
		uisignalkey = walk->data;
		if (uisignalkey->widget == widget)
			{
			ui_signaldisconnectbyfunc(uixml, widget,
			                          uisignalkey->parentcallback,
			                          uisignalkey);
			uixml->uisignalkeylist =
			    g_list_remove(uixml->uisignalkeylist, walk->data);
			mem_free(uisignalkey);
			return 1;
			}
		}
	return 0;
	}

void
ui_signalconnect_keypress_cleanup(UiXML * uixml)
	{
	UiSignalKey *uisignalkey;
	GList *walk;

	for (walk = g_list_first(uixml->uisignalkeylist); walk != NULL;
	        walk = walk->next)
		{
		uisignalkey = walk->data;
		ui_signalconnect_keypress_free(uixml, uisignalkey->widget);
		}

	if (!uixml->uisignalkeylist)
		return ;
	g_list_free(uixml->uisignalkeylist);
	}

void
ui_signalemitstopbyname(UiXML * uixml, UiWidget * object, const gchar * name)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * object,
	                          const gchar * name);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "signalemitstopbyname");
	func = f->ptr;

	func(uixml, object, name);
	}

void
ui_signalemitbyname_arg_pUiEvent_pgboolean(UiXML * uixml, UiWidget * object,
        const gchar * name, UiEvent * event,
        gboolean * retval)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * object,
	                          const gchar * name, UiEvent * event,
	                          gboolean * retval);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "signalemitbyname_arg_pUiEvent_pgboolean");
	func = f->ptr;

	func(uixml, object, name, event, retval);
	}

void
ui_event_handlekeyevent(UiXML * uixml, UiEvent * event, void *data)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiEvent * event, void *data);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "event_handlekeyevent");
	func = f->ptr;

	func(uixml, event, data);
	}

void
ui_event_setvalue(UiXML * uixml, UiEvent * event, UiEventType type,
                  UiEventValue value)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiEvent * event, UiEventType type,
	                          UiEventType value);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "event_setvalue");
	func = f->ptr;

	func(uixml, event, type, value);
	}

void
ui_editable_inserttext(UiXML * uixml, UiWidget * editable,
                       const gchar * new_text, gint new_text_length,
                       gint * position)
	{
	PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * editable,
	                          const gchar * new_text, gint new_text_length,
	                          gint * position);

	function func;

	f = plugin_findfunction(uixml, "editable_inserttext");
	func = f->ptr;

	func(uixml, editable, new_text, new_text_length, position);
	}

void
ui_editable_deletetext(UiXML * uixml, UiWidget * editable, gint start_pos,
                       gint end_pos)
	{
	PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * editable, gint start_pos,
	                          gint end_pos);

	function func;

	f = plugin_findfunction(uixml, "editable_deletetext");
	func = f->ptr;

	func(uixml, editable, start_pos, end_pos);
	}

void
ui_editable_setposition(UiXML * uixml, UiWidget * editable, gint position)
	{
	PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * editable, gint position);

	function func;

	f = plugin_findfunction(uixml, "editable_setposition");
	func = f->ptr;

	func(uixml, editable, position);
	}

void
ui_togglebutton_setactive(UiXML * uixml, UiWidget * toggle_button,
                          gboolean is_active)
	{
	PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget * editable,
	                          gboolean is_active);

	function func;

	f = plugin_findfunction(uixml, "togglebutton_setactive");
	func = f->ptr;

	func(uixml, toggle_button, is_active);
	}

/*
 * ui_list_getrowpos:
 *
 * Francis: This function shall be called by biglist_getrow()
 * 
 */

gint
ui_list_getrowpos(UiXML * uixml, UiWidget * widget)
	{
	static PluginFunction *f;
	typedef gint(*function) (UiXML * uixml, UiWidget *);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "list_getrowpos");
	func = f->ptr;

	return func(uixml, widget);
	}

void
ui_list_setrowpos(UiXML * uixml, UiWidget * widget, gint pos)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget *, gint);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "list_setrowpos");
	func = f->ptr;

	func(uixml, widget, pos);
	}

UiWidget *
ui_getwidgetinparent(UiXML * uixml, gchar * name, gchar * parentname)
	{
	UiWidget *widget = NULL;
	GList *children;

	children = ui_widgetchildren(uixml, parentname);

	while (children != NULL)
		{
		if (strcmp(name, ui_getname_short(uixml, children->data)) == 0)
			{
			widget = children->data;
			break;
			}
		children = children->next;
		}
	return widget;
	}

/**
 * ui_findgladefile:
 * @gladefile: Name of glade file to look for.
 * 
 * Look really hard for the glade file. Assuming it could be in /usr/share or 
 * /usr/local/share.
 */
gchar *
ui_findxmlfile(gchar * gladefile)
	{
	FILE *file;
	gint i, count;
	glob_t g;
	gchar *pattern, *dir;

	g_assert(gladefile);
	if ((file = fopen(gladefile, "r")) != NULL)
		{
		fclose(file);
		return gladefile;
		}

	dir = (gchar *) mem_alloc(strlen(gladefile) + 2);
	dir[0] = 0;
	count = strlen(gladefile);
	for (i = 0; i < count; i++)
		{
		if (gladefile[i] == '.')			 /* warning, file.file.glade is not
																	    supported. */
			break;
		dir[i] = gladefile[i];
		}
	dir[i] = 0;

	for (i = 0; i < NUMGLADESEARCH; i++)
		{
		pattern =
		    mem_strdup_printf("%s%s/%s", gladefilesearch[i], dir, gladefile);

		/* debugmsg("searching for %s", pattern); */
		if (glob(pattern, 0, NULL, &g) == 0)
			for (count = 0; count < g.gl_pathc; count++)
				if ((file = fopen(g.gl_pathv[count], "r")) != NULL)
					{
					globfree(&g);
					fclose(file);
					mem_free(dir);
					return pattern;
					}
		mem_free(pattern);
		}
	errormsg("Unable to find xml file %s anywhere", gladefile);
	mem_free(dir);
	globfree(&g);
	return NULL;
	}

/**
 * assign_position:
 * @widgetname:
 * @x:
 * @y:
 *  
 * Specify the x and y conordinates to display a form at.
 * Should only be called on forms, I haven't checked to see what
 * happens if you do it to other widgets - nothing exciting I imagine.
 */

void
ui_assign_position(UiXML * uixml, gchar * widgetname, gint x, gint y)
	{
	UiWidget *widget = ui_getwidget(uixml, widgetname);
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget *, gint, gint);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "assignpos");
	func = f->ptr;
	func(uixml, widget, x, y);
	}

/**
 * ui_get_selected_row_number
 * 
 * Given a clist, return the currently selected row.
 */

gint
ui_get_selected_row_number(UiXML * uixml, UiWidget * clist)
	{
	static PluginFunction *f;
	typedef gint(*function) (UiXML * uixml, UiWidget *);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "list_getselectedrownum");
	func = f->ptr;
	return func(uixml, clist);
	}

/**
 * ui_seteditable:
 *
 * This function allows you to disable widgets when you dont want them. This is done
 * if there is a zero records in a datasource.
 */
void
ui_seteditable(UiXML * uixml, UiWidget * widget, gboolean state)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget *, gboolean);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "seteditable");
	func = f->ptr;
	func(uixml, widget, state);
	}

/**
 * ui_geteditable:
 *
 * This function allows you to disable widgets when you dont want them. This is done
 * if there is a zero records in a datasource.
 */
gint
ui_geteditable(UiXML * uixml, UiWidget * widget)
	{
	static PluginFunction *f;
	typedef gint(*function) (UiXML * uixml, UiWidget *);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "geteditable");
	func = f->ptr;
	return func(uixml, widget);
	}

gchar *
ui_event_getstringfromevent(UiXML * uixml, UiWidget * widget)
	{
	static PluginFunction *f;
	typedef gchar *(*function) (UiXML * uixml, UiEvent *);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "event_getstringfromevent");
	func = f->ptr;
	return func(uixml, widget);
	}

void
ui_event_setstringfromevent(UiXML * uixml, UiWidget * widget, gchar * str)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiEvent *, gchar *);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "event_setstringfromevent");
	func = f->ptr;
	func(uixml, widget, str);
	}

/**
 * ui_setsensitive:
 * @widget: An widget to set sensitive or not
 * @val: True or false
 **/
void
ui_setsensitive(UiXML * uixml, UiWidget * widget, gboolean val)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget *, gboolean);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "setsensitive");
	func = f->ptr;
	func(uixml, widget, val);
	}

void
ui_list_append(UiXML * uixml, UiWidget * widget, GList * values,
               gulong secretid, gboolean append)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget *, GList *, gulong,
	                          gboolean);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "list_append");
	func = f->ptr;
	func(uixml, widget, values, secretid, append);
	}

void
ui_list_clear(UiXML * uixml, UiWidget * widget)
	{
	static PluginFunction *f;
	typedef void (*function) (UiXML * uixml, UiWidget *);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "list_clear");
	func = f->ptr;
	func(uixml, widget);
	}

gboolean
ui_list_setselection(UiXML * uixml, UiWidget * widget, gulong secretid)
	{
	static PluginFunction *f;
	typedef gboolean(*function) (UiXML * uixml, UiWidget *, gulong);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "list_setselection");
	func = f->ptr;
	return func(uixml, widget, secretid);
	}

gulong
ui_list_getselection(UiXML * uixml, UiWidget * widget, gboolean getvalues,
                     GList ** values)
	{
	static PluginFunction *f;
	typedef gulong(*function) (UiXML * uixml, UiWidget *, gboolean, GList **);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "list_getselection");
	func = f->ptr;
	return func(uixml, widget, getvalues, values);
	}

gint
ui_list_setcellvalue(UiXML * uixml, UiWidget * widget, gint col,
                     gulong secretid, gchar * value)
	{
	static PluginFunction *f;
	typedef gulong(*function) (UiXML * uixml, UiWidget *, gint, gulong, gchar *);
	function func;

	if (!f)
		f = plugin_findfunction(uixml, "list_setcellvalue");
	if (f != NULL)
		{
		func = f->ptr;
		return func(uixml, widget, col, secretid, value);
		}
	return -1;
	}

gint
ui_display_message(UiXML * uixml, gchar * message, gchar * title, gint type)
	{
	static PluginFunction *f;
	typedef gint(*function) (gchar * message, gchar *title, gint type);
	function func;

	if (!f)
		if ((f = plugin_findfunction(uixml, "message_display")) == NULL)
			return -1;

	func = f->ptr;
	return func(message, title, type);
	}


gint
ui_raisewindow(UiXML * uixml, UiWidget *widget)
	{
	return (gint)plugin_exec_arg1(uixml, "raisewindow", widget);
	}
