/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <glob.h>
#include <dlfcn.h>
#include <string.h>
#include <glib.h>

/* Lots of includes here */
#include "uiplugin.h"
#include "bc.h"
#include "biglists.h"

/* Where to search for the plugin */
#define NUMPLUGINPATHS 8
char *plugin_search_path[NUMPLUGINPATHS] = {
        "../uigtk/",
        "../uigtk/.libs/",
        "../uihtml/.libs/",
        "../uiwxwin/.libs/",
        "./",
        "./.libs/",
        "/usr/local/lib/",
        "/usr/lib/"
                                           };

/* load that library eh */
static void *
check_shared_library(char *so)
	{
	void *handle;
	void *sym;

	if ((handle = dlopen(so, RTLD_LAZY)) == 0)
		{
		errormsg("An error occured while reading the shared library "
			"%s\n%s", so, dlerror());
		return NULL;
		}

	if ((sym = dlsym(handle, "plugin_init")) == 0)
		{
		dlclose(handle);
		errormsg("Unable to find the plugin_init function");
		return NULL;
		}
	return handle;
	}

/**
 * plugin_free:
 *
 *
 * 
 * Returns: Can't returning anything
 */

void
plugin_free(Plugin * plugin)
	{
	g_assert(plugin);
	
	dlclose(plugin->dlhandle);
	
	mem_free(plugin->name);
	mem_free(plugin->args);
	mem_free(plugin->func_gridedit->name);
	mem_free(plugin->func_gridedit);
	mem_free(plugin->func_blockemit->name);
	mem_free(plugin->func_blockemit);
	
	mem_free(plugin);
	}

/* For handling calls back into bond from the ui plugin */
static void
plugin_add_bondexportfunctions_set(PluginFunction *func, gchar * funcname, 
		void *(*funcptr) ())
	{
	func->name = mem_strdup(funcname);
	func->ptr = funcptr;	
	}

		
static void
plugin_add_bondexportfunctions(Plugin *plugin)
	{
	plugin->func_gridedit = mem_calloc(sizeof(PluginFunction));
	plugin_add_bondexportfunctions_set(plugin->func_gridedit,
			"gridedit",(void*)biglist_sig_gridedited);
	plugin->func_blockemit = mem_calloc(sizeof(PluginFunction));
	plugin_add_bondexportfunctions_set(plugin->func_blockemit,
			"blockemit",(void*)postit_sig_blockemit);
	}

/**
 * plugin_create:
 * @library: Name of a library to load
 * 
 * Loook for a library and create a plugin for it.  
 * 
 * Returns: a Plugin.
 */
Plugin *
plugin_create(gchar * library, gchar * args)
	{
	Plugin *plugin = NULL;
	gint i, count;
	glob_t g;
	gchar *pattern;

	g_assert(library);

	plugin = mem_alloc(sizeof(Plugin));
	if (args)
		args = mem_strdup(args);
	if (library)
		library = mem_strdup(library);

	plugin->name = library;
	plugin->args = args;
	plugin->func = NULL;
	plugin->numfunctions = 0;

	plugin_add_bondexportfunctions(plugin);
	
	for (i = 0; i < NUMPLUGINPATHS; i++)
		{
		pattern = mem_strdup_printf("%slibbond_ui*.so", 
				plugin_search_path[i]);

		if (glob(pattern, 0, NULL, &g) == 0)
			{
			for (count = 0; count < g.gl_pathc; count++)
				{
				debug_output("Checking library: %s\n", 
					g.gl_pathv[count]);

				plugin->dlhandle = 
					check_shared_library(g.gl_pathv[count]);
				
				if (plugin->dlhandle != NULL)
					{
					plugin->sym = dlsym(plugin->dlhandle, 
							"plugin_init");
					g_assert(plugin->sym);
					if (plugin->sym(plugin) == 0)
						{
						globfree(&g);
						mem_free(pattern);
						debug_output("Plugin "
						"successfully loaded.\n");
						return plugin;
						}
					}
				}
			}
		mem_free(pattern);
		}
	errormsg("Failed to load plugin library for %s\n", library);
	mem_free(plugin->name);
	mem_free(plugin->args);
	mem_free(plugin);

	return NULL;
	}

/**
 * plugin_findfunction:
 * @plugin: a plugin created by plugin_create()
 * @name: name of function to load and return.
 *
 *
 * Returns: A plugin function else %NULL on error.
 */
PluginFunction *
plugin_findfunction(UiXML *uixml, gchar * name)
	{
	static gchar *lastname;
	static PluginFunction *lastreturn;
	gint i;

	if (lastname && strcmp(name,lastname)==0)
		return lastreturn;
	
	g_assert(uixml->plugin);
		
	for (i = 0; i < uixml->plugin->numfunctions; i++)
		{
		if (strcmp(name, uixml->plugin->func[i]->name) == 0)
			{
			lastname = name;
			lastreturn = uixml->plugin->func[i];
			return lastreturn;
			}
		/* todo, add more impressive searching */
		}
	errormsg("Unable to find function %s", name);
	return NULL;
	}

/**
 * plugin_exec:
 * @plugin: 
 * @funcname: function name to excute
 *
 * Find the @funcname in the plugin library and then excute it with
 * zero argurements.  It assumes the format int funcname(void);
 *
 * Returns: whatever the other function returned.
 */

gint
plugin_exec(UiXML *uixml, gchar * funcname)
	{
	PluginFunction *f;
	typedef gint(*function) (UiXML *);
	function func;

	if ((f = plugin_findfunction(uixml, funcname)) == NULL)
		return -1;
	func = f->ptr;
	return func(uixml);
	}

/**
 * plugin_exec_arg1:
 * @plugin: plugin library.
 * @funcname: Function to call
 * @arg1: void* to pass in as argument one
 *
 * Returns: %NULL on error else a void* from the other function
 */
void *
plugin_exec_arg1(UiXML *uixml, gchar * funcname, void *arg1)
	{
	PluginFunction *f;
	typedef void *(*function) (UiXML *, void *);
	function func;

	if ((f = plugin_findfunction(uixml, funcname)) == NULL)
		return NULL;
	func = f->ptr;
	return func(uixml,arg1);
	}

void *
plugin_exec_arg2(UiXML *uixml, gchar * funcname, void *arg1, void *arg2)
	{
	PluginFunction *f;
	typedef void *(*function) (UiXML *, void *, void *);
	function func;

	if ((f = plugin_findfunction(uixml, funcname)) == NULL)
		return NULL;
	func = f->ptr;
	return func(uixml,arg1, arg2);
	}
