#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "bc.h"
#include "uihtml.h"
#include "uibox.h"
#include "uigeneral.h"

/**
 * uihtml_box_create:
 * @widgetnode: an xml node in the bond XML file.
 * @parentwidget: widget that will be the parent of this.
 *
 * Creates a table/box based on properties specified in the XML
 * @node. It makes the widget visible by default.
 *
 * Attributes:
 * name:           Specifyies the name to be given to this widget.
 * columns:        Number of columns.
 * rows:           Number of rows.
 * homogeneous:    Specifies whether cells should all be same size. Can
 *                 be either "true" or "false".
 * visible:        "true" or "false".
 * column_spacing: Number pixels between each column.
 * row_spacing:    Number pixels between each row.
 * spacing:        Number pixels between each column and row. If this
 *                 attribute is used, will override column_spacing and
 *                 row_spacing.
 *
 * Example:        <box name="givenname" columns="2" spacing="3">...</box>
 *
 * Returns: Pointer to the widget created.
 */
UiHtml *
uihtml_box_create(UiXML * uixml, xmlNode * node, UiHtml * parentwidget,
		 gboolean * visible)
{
	UiHtml *box;
	xmlChar *content;
	/*guint columns = 1, rows = 1;*/
	guint column_spacing = 4, row_spacing = 4;
	gboolean homogeneous = FALSE;
	
	g_assert(node);
	box = uihtml_create(node);
	box->widgettype = UiGeneral;
	box->openinghtml = mem_strdup("<table class=\"box\">");

	content = xmlGetProp(node, "homogeneous");
	if (content && (IsTrue(content)))
	  homogeneous = TRUE;
	
	content = xmlGetProp(node, "visible");
	if (content && IsFalse(content))
	  *visible = FALSE;
	else *visible = TRUE;
	
	content = xmlGetProp(node, "column_spacing");
	if (content)
	  column_spacing = atoi(content);
	
	content = xmlGetProp(node, "row_spacing");
	if (content)
	  row_spacing = atoi(content);

	content = xmlGetProp(node, "columns");
	if (content)
	  box->xsize = atoi(content);

	content = xmlGetProp(node, "rows");
	if (content)
	  box->ysize = atoi(content);

	content = xmlGetProp(node, "spacing");
	if ((content) && (column_spacing = atoi(content))) {
	  row_spacing = column_spacing;
	}

	box->closinghtml = mem_strdup("</table><!-- box -->");

	debuglog(50, "Creating box %dx%d", box->xsize, box->ysize);
	
	return box;
}


/**
 * uihtml_box_add:
 * @widgetnode: an xml node in the bond XML file.
 * @parentwidget: widget that will be the parent of this.
 *
 * Adds a cell to a table/box based on properties specified in the XML
 * @node.
 *
 * Options:
 * xpos: column position in table.
 * ypos: row position in table
 * xpad:
 * ypad:
 * xspan: span of cell
 * yspan: span of cell
 * xexpand:
 * yexpand:
 * xfill:
 * yfill:
 * xshrink:
 * yshrink:
 *
 * Example:
 * <cell ypos="1">
 ...
 * </cell>
 *
 * Returns: Pointer to the widget created.
 */

gint
uihtml_box_add(xmlNode * node, UiHtml * grandchild, UiHtml * parentwidget)
{
	xmlChar *content;
	gint xpad = -1, ypad = -1, xspan, yspan, xpos, ypos;
	UiHtml * cell;

	if (GTK_IS_WIDGET(grandchild) || GTK_IS_WIDGET(parentwidget))
	  {
		errormsg("Invalid widgeT Line: %d\n",xmlGetLineNo(node));
		return -1;
	  }
	cell = uihtml_create(node);
	cell->widgettype = UiGeneral;

	mem_verify(parentwidget);
	mem_verify(grandchild);
	cell->children = g_list_append(cell->children,grandchild);
	debugmsg("Grandchild: %d cell %d",grandchild->widgettype,cell->widgettype);
	parentwidget->children = g_list_append(parentwidget->children,cell);
	cell->parent = parentwidget;
	grandchild->parent = cell;
	
	cell->openinghtml = mem_strdup("<td class=\"box-cell\""); 
	cell->closinghtml = mem_strdup("</td>");
	
	content = xmlGetProp(node, "xpos");
	if (content)
	  {
		xpos = atoi(content);
		if(xpos == 0)
		  cell->openinghtml = mem_strdup_printf("<tr>%s", cell->openinghtml);
		if(xpos == (parentwidget->xsize - 1))
		  cell->closinghtml = mem_strdup_printf("%s</tr>", cell->closinghtml);
	  }
	
	content = xmlGetProp(node, "ypos");
	if (content && (ypos = atoi(content)))
	  {
		;
	  }
	
	content = xmlGetProp(node, "xspan");
	if (content && (xspan = atoi(content)))
	  cell->openinghtml = 
		mem_strdup_printf("%s colspan=\"%d\"", cell->openinghtml, xspan);

	content = xmlGetProp(node, "yspan");
	if (content && (yspan = atoi(content)))
	  cell->openinghtml = 
		mem_strdup_printf("%s rowspan=\"%d\"", cell->openinghtml, yspan);
	
	content = xmlGetProp(node, "xpad");
	if (content)
	  xpad = atoi(content);

	content = xmlGetProp(node, "ypad");
	if (content)
	  ypad = atoi(content);
	
	content = xmlGetProp(node, "padding");
	if (content) {
	  ypad = atoi(content); xpad = ypad;
	}

	if((xpad >= 0) || (ypad >= 0))
	  {
		cell->openinghtml = mem_strdup_printf("%s style=\"", 
											  cell->openinghtml);
		if(xpad==ypad)
		  cell->openinghtml = 
			mem_strdup_printf("%spadding:%d;",
							  cell->openinghtml, xpad);
		else
		  {
			if(xpad >= 0)
			  cell->openinghtml = 
				mem_strdup_printf("%spadding-left:%d;padding-right:%d;",
								  cell->openinghtml, xpad, xpad);
			if(ypad >= 0)
			  cell->openinghtml = 
				mem_strdup_printf("%spadding-top:%d;padding-bottom:%d",
								  cell->openinghtml, ypad, ypad);
		  }
		cell->openinghtml = mem_strdup_printf("%s\"", 
											  cell->openinghtml);
	  }
	
	cell->openinghtml = mem_strdup_printf("%s >", cell->openinghtml);

	content = xmlGetProp(node, "xexpand");
	if (content && IsFalse(content))
	  ;
		/*xoptions &= ~GTK_EXPAND;*/


	content = xmlGetProp(node, "yexpand");
	if (content) {
		if (IsTrue(content))
		  /*yoptions |= GTK_EXPAND*/
		  ;
		else if (IsFalse(content))
			/*yoptions &= ~GTK_EXPAND*/
		  ;
	}

	content = xmlGetProp(node, "expand");
	if (content && IsFalse(content)) {
		/*xoptions &= ~GTK_EXPAND;
		yoptions |= GTK_EXPAND*/
	  ;
	}

	content = xmlGetProp(node, "xfill");
	if (content && IsFalse(content))
		/*xoptions &= ~GTK_FILL*/
	  ;

	content = xmlGetProp(node, "yfill");
	if (content && IsFalse(content))
		/*yoptions &= ~GTK_FILL*/
	  ;

	content = xmlGetProp(node, "fill");
	if (content && IsFalse(content)) {
		/*xoptions &= ~GTK_FILL;
		yoptions &= ~GTK_FILL*/
	  ;
	}
	
	cell->order = (parentwidget->xsize * (ypos+1)) + (xpos+1);

/*	cell->openinghtml = mem_strdup_printf("%s order:%d", 
										  cell->openinghtml, cell->order);
*/
	debugmsg("**** creating cell with order [%d]x=%d, y=%d", cell->order, xpos, ypos);
	
	return 0;
}

UiHtml *
uihtml_vhbox_create(UiXML * uixml, xmlNode * widgetnode, UiHtml * parentwidget,
				   gboolean vboxtrue, gboolean * visible)
{
	xmlChar *content;
   	UiHtml *vhbox;
	gint spacing = 4;
	
  if (uixml->defaultvisible)
	*visible = TRUE;
  else
	*visible = FALSE;
  
  content = xmlGetProp(widgetnode, "visible");
  if (content && IsFalse(content))
	*visible = FALSE;
  else if(content && IsTrue(content)) 
	*visible = TRUE;

	content = xmlGetProp(widgetnode, "spacing");
	if (content) {
		spacing = atoi(content);
		if (spacing < 0)
			spacing = 1;
	}

	debuglog(53, "Creating a box of some kind.");
   
   	vhbox = uihtml_create(widgetnode);
   	vhbox->visible = *visible;
	vhbox->vhordercounter=0;
	
	
	if (vboxtrue == 1)
	  {
		vhbox->vboxtrue = 1;
		vhbox->openinghtml = mem_strdup("<table class=\"vbox\"");
		vhbox->closinghtml = mem_strdup("</table><!-- vbox -->");
	  }
	else
	  {
		vhbox->vboxtrue = 0;
		vhbox->openinghtml = mem_strdup("<table class=\"hbox\"");
		vhbox->closinghtml = mem_strdup("</table><!-- hbox -->");
	  }
	vhbox->openinghtml = mem_strdup_printf("%s>", vhbox->openinghtml);
   		
   	return vhbox;
}

gint 
uihtml_vhbox_add(xmlNode * node, UiHtml * grandchild, UiHtml * parentwidget)
{
	xmlChar *content;
	gint padding = -1;
	UiHtml * cell;

	if (GTK_IS_WIDGET(grandchild) || GTK_IS_WIDGET(parentwidget))
	  {
		errormsg("Invalid widgeT Line: %d\n",xmlGetLineNo(node));
		return -1;
	  }
	cell = uihtml_create(node);
	cell->widgettype = UiGeneral;

	mem_verify(parentwidget);
	mem_verify(grandchild);
	cell->children = g_list_append(cell->children,grandchild);
	debugmsg("Grandchild: %d cell %d\n",grandchild->widgettype,cell->widgettype);
	parentwidget->children = g_list_append(parentwidget->children,cell);
	cell->parent = parentwidget;
	grandchild->parent = cell;
	
	cell->order = parentwidget->vhordercounter;
	parentwidget->vhordercounter ++;

	content = xmlGetProp(node, "padding");
	if (content)
	  padding = atoi(content);
	
	cell->openinghtml = mem_strdup("");
	
	if(parentwidget->vboxtrue)
	  /*vbox*/
	  {
		cell->openinghtml = mem_strdup("<tr><td class=\"vbox-cell\"");
		cell->closinghtml = mem_strdup("</td></tr>");
	  }
	else
	  /*hbox*/
	  {
/*		printf("i am an hbox cell, my order is %d\n", cell->order);*/
		/* if first cell*/	
		if(cell->order == 0)
		  cell->openinghtml = mem_strdup("<tr>");
		
		cell->openinghtml = 
		  mem_strdup_printf("%s<td class=\"hbox-cell\"", cell->openinghtml);
		
		cell->closinghtml = mem_strdup("</td>");
		
		/* if last cell -- TO DO (also in uihtml_toolbar_add)*/
		if(0)
		  cell->closinghtml = mem_strdup("</tr>");
	  }

	if(padding >= 0)
	  cell->openinghtml = mem_strdup_printf("%s style=\"padding:%d;\"", 
											cell->openinghtml, padding);
	
	cell->openinghtml = mem_strdup_printf("%s>", cell->openinghtml);
	
	debuglog(50, "Adding widget to vhbox. Yes it did happen");

	return 0;
}
