#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>

#include <glib.h>

#include "wssocket.h"
#include "debug.h"
#include "wsmem.h"

#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))

/* how long to wait before reporting an error in socket_recv */
#define SOCKET_TIMEOUT 60*30

struct asocket
{
	int sockfd;
	struct sockaddr_in addr;
};

/* Sets up a socket for listening on port port
 */

asocket *
socket_listen(int port)
{
	asocket *retval;

	retval = mem_alloc(sizeof(asocket), "socket_listen");
	if (!retval)
		return NULL;

	if ((retval->sockfd = socket(PF_INET, SOCK_STREAM, 0)) == -1)
		{
			mem_free(retval);
			perror("socket error");
			return NULL;
		}

	/* tracer_log("Got sockfd: %d", retval->sockfd); */
	retval->addr.sin_family = AF_INET;
	retval->addr.sin_port = htons(port);
	retval->addr.sin_addr.s_addr = INADDR_ANY;
	memset(&retval->addr.sin_zero, 0, 8);

	if (bind(retval->sockfd, (struct sockaddr *)&retval->addr, sizeof(struct sockaddr)) == -1)
		{
			mem_free(retval);
			perror("bind error");
			return NULL;
		}

	if (listen(retval->sockfd, 10) == -1)
		{
			perror("listen error");
			mem_free(retval);
			return NULL;
		}

	return retval;
}

/* waits for a new connection to an already listening socket
 * mallocs memory for the return value (which should later be free'd
 * using socket_destroy)
 */
asocket *
socket_accept(asocket * socket)
{
	socklen_t sin_size;
	asocket *retval;
	char buf[32];
	static int count = 0;
	int opt;
	char *info;

	sin_size = sizeof(struct sockaddr_in);

	count++;
	sprintf(buf, "socket_accept(%d)", count);
	retval = mem_alloc(sizeof(asocket), buf);

	if ((retval->sockfd = accept(socket->sockfd, (void *)&retval->addr, &sin_size)) == -1)
		{
			mem_free(retval);
			perror("accept error");
			return NULL;
		}
	opt = 1;
	if (setsockopt(retval->sockfd, SOL_TCP, TCP_NODELAY, &opt, sizeof(opt)) == -1)
		perror("setsockopt error");

	opt = 1;
	if (setsockopt(retval->sockfd, SOL_SOCKET, SO_KEEPALIVE, &opt, sizeof(opt)) == -1)
		perror("setsocketopt error,keepalive.");

	info = socket_info(retval);

	debugmsg("Accepted connection from %s on sockfd %d", info, retval->sockfd);

	return retval;
}

asocket *
socket_connect(char *host, int port)
{
	asocket *retval;
	struct hostent *he;

	if ((he = gethostbyname(host)) == NULL)
		return NULL;

	retval = malloc (sizeof *retval);

	if ((retval->sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		{
		free (retval);
		return NULL;
		}

	retval->addr.sin_family = AF_INET;			 // host byte order 
	retval->addr.sin_port = htons(port);		 // short, network byte order 
	retval->addr.sin_addr = *((struct in_addr *)he->h_addr);
	memset(&(retval->addr.sin_zero), '\0', 8);	// zero the rest of the struct 

	if (connect(retval->sockfd, (struct sockaddr *)&retval->addr, sizeof(struct sockaddr)) == -1)
		{
		free (retval);
		return NULL;
		}

	return retval;
}

/* sends a block of data down a socket. Every MAXPACKETLEN bytes, it waits
 * to receive a single byte containing the number 1 (0x01), if this is not
 * received then there is an error
 */
int
socket_send(asocket * socket, void *buffer, int len)
{
	int sendlen;

	if (!socket)
		return -1;

	sendlen = send(socket->sockfd, buffer, len, 0);

	if (sendlen < 0)
		{
			perror("send");
			return -1;
		}

	if (sendlen != len)
		return -1;

	return 0;
}

/* receives data from an existing socket. every MAXPACKETLEN bytes it will
 * send a single bytes containing '1' (0x01)
 */
int
socket_recv(asocket * socket, void *buffer, int len)
{
	int recvlen;

	/* tracer_log( "expecting %d bytes" , len ); */

	if (!socket)
		return -1;

	recvlen = recv(socket->sockfd, buffer, len, 0);
	if (recvlen < 0)
		{
			perror("recv");
			return -1;
		}

	if (recvlen != len)
		{
			debugmsg("recieved %d instead of %d", recvlen, len);

			if (recvlen == 1)
				debugmsg("got: %d\n", ((char *)buffer)[0]);

			debugmsg("buffer: %s\n", (char *)buffer);
			return -1;
		}

	return 0;
}

/* free's all resources associated with a socket */
int
socket_destroy(asocket * socket)
{
	if (!socket)
		return 0;
	shutdown (socket->sockfd, SHUT_RDWR);
	close(socket->sockfd);
	mem_free(socket);
	return 0;
}

/* returns some info about a socket, atm this is its IP address in string
 * format
 */
char *
socket_info(asocket * socket)
{
	if (!socket)
		return "Invalid socket";

	return inet_ntoa(socket->addr.sin_addr);
}

int
socket_address(asocket * socket)
{
	return inet_lnaof(socket->addr.sin_addr);
}

char *
socket_recvline(asocket * socket)
{
	char *retval;
	int count = 0;
	int done = 0;
	int maxsize = 64;

	retval = malloc(maxsize);

	while (!done)
		{
			if (socket_recv(socket, &retval[count], 1) == -1)
				{
					count--;
					done = 1;
				}
			if (retval[count] == '\n')
				{
					retval[count - 1] = retval[count] = '\0';
					done = 1;
				}
			count++;

			if (count >= maxsize - 2)
				{
					maxsize += 64;
					retval = realloc(retval, maxsize);
				}
		}

	retval[count] = '\0';

	return retval;
}

int
socket_sendline(asocket * socket, char *line)
{
	int retval;

	retval = socket_send(socket, line, strlen(line));

	if (retval != 0)
		return retval;

	retval = socket_send(socket, "\r\n", 2);

	return retval;
}

int socket_fd (asocket *socket)
{
	if (!socket)
		return 0;

	return socket->sockfd;
}