#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <libxml/parser.h>
#include <glib.h>

#include "uihtml.h"
#include "debug.h"
#include "uihtmlinit.h"
#include "uihtmldraw.h"

#define output_html(fmt,args...) fprintf (output, "%*s" fmt , indent * 2, "" , ##args)

static xmlNode *
get_node (xmlNode *node, char *name)
{
	xmlNode *child;

	if (!node)
		return NULL;

/*	debugmsg ("Searching for %s node in %s", name, node->name); */

	for (child = node->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, name) == 0)
			return child;
		}

	return NULL;
}

static char *
get_node_content (xmlNode *node, char *name)
{
	xmlNode *child;
	child = get_node (node, name);
	if (child)
		return child->content;
	return NULL;
}

static xmlNode *
get_textnode (xmlNode *widget)
{
	xmlNode *child;

	for (child = widget->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "text") == 0 && child->type == XML_ELEMENT_NODE)
			return child;
		}

	return NULL;
}

static gint
get_padding (xmlNode *widget)
{
	int retval;
	xmlNode *layout, *padding;

	if (!widget)
		return 0;

	layout = get_node (widget, "child");

	if (!layout)
		return 0;

	padding = get_node (widget, "padding");

	if (!padding || !padding->content)
		return 0;

	retval = atoi (padding->content);

	return retval;
}

static gint
draw_label (FILE *output, int indent, xmlNode *label)
{
	char *text = get_node_content (get_node (label, "label"), "text");

	output_html ("%s\r\n", text);

	return 0;
}

static gint
draw_text (FILE *output, int indent, xmlNode *textnode)
{
	char *name, *text;

	name = get_node_content (get_node (textnode, "name"), "text");
	text = get_node_content (get_textnode (textnode), "text");

	output_html ("<textarea name=\"%s\">", name);

	if (text)
		fprintf (output, "%s\r\n", text);

	fprintf (output, "</textarea>\r\n");

	return 0;
}

static gint
draw_entry (FILE *output, int indent, xmlNode *entry)
{
	char *name, *text;

	name = get_node_content (get_node (entry, "name"), "text");
	text = get_node_content (get_textnode (entry), "text");

	output_html ("<input name=\"%s\" type=\"text\"", name);
	if (text)
		fprintf (output, " value=\"%s\"", text);
	fprintf (output, ">\r\n");

	return 0;
}

static gint
draw_spin (FILE *output, int indent, xmlNode *spin)
{
	char *name;
	name = get_node_content (get_node (spin, "name"), "text");
	output_html ("<input name=\"%s\" type=\"text\">\r\n", name);
	return 0;
}

static gint
draw_empty (FILE *output, int indent, xmlNode *empty)
{
	xmlNode *child;

	for (child = empty->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			uihtml_draw_widget (output, indent + 1, child);
		}

	return 0;
}

static gint
draw_combo (FILE *output, int indent, xmlNode *combo)
{
	char *items, *pos, *cur, *name;

	name = get_node_content (get_node (combo, "name"), "text");

	output_html ("<select name=\"%s\">\r\n", name);

	items = get_node_content (get_node (combo, "items"), "text");

	if (items)
		{
		do
			{
			pos = strchr (items, '\n');

			if (pos)
				{
				asprintf (&cur, "%s", items);
				cur[pos - items] = '\0';
				items = pos+1;
				}
			else
				cur = strdup (items);

			output_html ("<option>%s</option>\r\n", cur);
			free (cur);
			}
		while (pos && items && *items);
	}

	output_html ("</select>\r\n");

	return 0;
}

static gint
draw_button (FILE *output, int indent, xmlNode *button)
{
	char *label, *name;

	name = get_node_content (get_node (button, "name"), "text");
	label = get_node_content (get_node (button, "label"), "text");
		
	output_html ("<input TYPE=\"SUBMIT\" NAME=\"%s\" VALUE=\"%s\">\r\n",
				name, label);

	return 0;
}

static gint
draw_checkbutton (FILE *output, int indent, xmlNode *button)
{
	char *label, *name;

	name = get_node_content (get_node (button, "name"), "text");
	label = get_node_content (get_node (button, "label"), "text");

	output_html ("<input type=\"checkbox\" name=\"%s\" style=\"width: auto;\" id=\"%s\"><label for=\"%s\">%s</label>\r\n", name, name, name, label);

	return 0;
}

static gint
draw_radiobutton (FILE *output, int indent, xmlNode *button)
{
	char *class, *parentname, *label;
	/*<input type="radio" name="
    nickname" value="No" checked>No <input type="radio" name="
    nickname" value="Yes">Yes*/

	label = get_node_content (get_node (button, "label"), "text");
	parentname = get_node_content (get_node (button->parent, "name"), "text");
	asprintf (&class, "radio-%s", parentname);

	output_html ("<input type=\"radio\" name=\"%s\" style=\"width: auto;\" id=\"%s\"><label for=\"%s\">%s</label>\r\n", class, class, class, label);

	return 0;
}

static gint
draw_separator (FILE *output, int indent, xmlNode *separator)
{
	output_html ("<hr>\r\n");
	return 0;
}

static gint
draw_pixmap (FILE *output, int indent, xmlNode *pixmap)
{
	char *name;
	char *filename;

	name = get_node_content (get_node (pixmap, "name"), "text");
	filename = get_node_content (get_node (pixmap, "filename"), "text");

	output_html ("<img name=\"%s\" src=\"pixmaps/%s\">\r\n", name, filename);

	return 0;
}

static gint
draw_alignment (FILE *output, int indent, xmlNode *alignment)
{
	char *name;
	int printed = 0;
	xmlNode *child;

	name = get_node_content (get_node (alignment, "name"), "text");

	output_html ("<table> <!-- alignment %s -->\r\n", name);

	for (child = alignment->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			{
			if (printed)
				{
				errormsg ("alignment has more than one child");
				continue;
				}

			uihtml_draw_widget (output, indent+1, child);
			printed = 1;
			}
		}

	output_html ("</table>\r\n");

	return 0;
}

static gint
draw_clist (FILE *output, int indent, xmlNode *clist)
{
	xmlNode *child;
	xmlNode *labelnode;
	char *label, *title;
	int count = 0;
	int ncolumns=0;

	labelnode = get_node (clist, "name");
	title = get_node_content (labelnode, "text");

	output_html ("<table border=\"1\" cellpadding=\"0\"> <!-- CList: %s -->\r\n", title);

	ncolumns = atoi (get_node_content (get_node (clist, "columns"), "text"));

	output_html ("<TR>\r\n");
	for (child = clist->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			{
			labelnode = get_node (child, "child_name");
			label = get_node_content (labelnode, "text");
			if (strncmp (label, "CList:", 6) == 0)
				{
				output_html ("<TD><b>\r\n");
				draw_label (output, indent + 1, child);
/*				labelnode = get_node (child, "name");
				label = get_node_content (labelnode, "text"); */
				output_html ("</b></TD>\r\n");
				}
			}
		}
	output_html ("</TR>\r\n");

	for (child = clist->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			{
			labelnode = get_node (child, "child_name");
			label = get_node_content (labelnode, "text");
			if (strncmp (label, "CList:", 6) == 0)
				continue;

			if (count % ncolumns == 0)
				fprintf (output, "<TR>\r\n");

			labelnode = get_node (child, "name");
			label = get_node_content (labelnode, "text");
			output_html ("<TD><a href=\"#%s-%d\">%s</option></TD>\r\n", title, count / ncolumns, label);

			if (count % ncolumns == ncolumns - 1)
				output_html ("</TR>\r\n");

			count++;
			}
		}

	output_html ("</table>  <!-- /CList: %s -->\r\n", title);

	return 0;
}

#define get_offset(x,y) (((y) * columns) + (x))

static gint
draw_table (FILE *output, int indent, xmlNode *table)
{
	xmlNode **table_entries, *child, *node, *layout;
	int rows, columns, x, y, offset, count;
	int top, bottom, right, left;

	rows = atoi(get_node_content (get_node (table, "rows"), "text"));
	columns = atoi(get_node_content (get_node (table, "columns"), "text"));
	
	/* debugmsg ("rows: %d, columns: %d", rows, columns); */

	if (rows == 0 || columns == 0)
		{
		errormsg ("Unable to get rows or columns");
		return -1;
		}

	table_entries = calloc (rows * columns, sizeof (xmlNode *));
	if (!table_entries)
		{
		errormsg ("Unable to allocate memory");
		return -1;
		}
	count = 0;

	for (child = table->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			{
			layout = get_node (child, "child");
			if (!layout)
				{
				errormsg ("Node does not have layout!");
				continue;
				}

			top = atoi (get_node_content (get_node (layout, "top_attach"), "text"));
			left = atoi (get_node_content (get_node (layout, "left_attach"), "text"));
			bottom = atoi (get_node_content (get_node (layout, "bottom_attach"), "text"));
			right = atoi (get_node_content (get_node (layout, "right_attach"), "text"));

			/* debugmsg ("cur pos: %d, %d", top, left); */

			offset = get_offset (left, top);
			/* offset = top * rows + left; */

			/* debugmsg ("offset: %d (%d, %d) (%d, %d)", offset, rows, columns, top, left); */

			if (top < rows && left < columns)
				table_entries[offset] = child;
			else
				errormsg ("Invalid table position: %dx%d", top, left);

			for (y = 0; y < bottom - top; y++)
				for (x = 0; x < right - left; x++)
					{
					if (y == 0 && x == 0)
						continue;
					/*offset = (top + y) * rows + (left + x); */
					offset = get_offset (left + x, top + y);
					node = table_entries[offset];

					if (node)	
						errormsg ("This one is already assigned: %dx%d", top + y, left + x);
					else
						table_entries[offset] = (void *)1;
					}

			}
		}

	output_html ("<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\">\r\n");

	for (y = 0; y < rows; y++)
		{
		output_html ("<tr>\r\n");
		for (x = 0; x < columns; x++)
			{
			/*offset = (y * rows) + x;*/
			offset = get_offset (x, y);
			node = table_entries[offset];
			if (node == (void *)1)
				continue;
			if (node)
				{
				layout = get_node (node, "child");
				if (!layout)
					{
					errormsg ("No layout");
					continue;
					}

				bottom = atoi (get_node_content (get_node (layout, "bottom_attach"), "text"));
				top = atoi (get_node_content (get_node (layout, "top_attach"), "text"));
				right = atoi (get_node_content (get_node (layout, "right_attach"), "text"));
				left = atoi (get_node_content (get_node (layout, "left_attach"), "text"));
			
				output_html ("<td");
				if (bottom - top > 1)
					fprintf (output, " rowspan=\"%d\"", bottom - top);
				if (right - left > 1)
					fprintf (output, " colspan=\"%d\"", right - left);
				fprintf (output, ">\r\n");
				uihtml_draw_widget (output, indent + 1, node);
				}
			else
				output_html ("<td>\r\n");

			output_html ("</td>\r\n");
			}

		output_html ("</tr>\r\n");
		}
	
	output_html ("</table>\r\n");

	free (table_entries);

	return 0;
}

static gint
draw_hbox (FILE *output, int indent, xmlNode *hbox)
{
	xmlNode *child;
	char *name;

	name = get_node_content (get_node (hbox, "name"), "text");

	output_html ("<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\"> <!-- hbox %s -->\r\n", name);

	output_html ("<tr>\r\n");

	for (child = hbox->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			{
			output_html ("<td>\r\n");
			uihtml_draw_widget (output, indent + 1, child);
			output_html ("</td>\r\n");
			}
		}

	output_html ("</tr>\r\n");

	output_html ("</table>");

	return 0;
}

static gint
draw_vbox (FILE *output, int indent, xmlNode *vbox)
{
	xmlNode *child;
	char *name;

	name = get_node_content (get_node (vbox, "name"), "text");

	output_html ("<TABLE> <!-- vbox %s -->\r\n", name);

	for (child = vbox->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			{
			output_html ("<TR>\r\n");
			uihtml_draw_widget (output, indent + 1, child);
			output_html ("</TR>\r\n");
			}
		}

	output_html ("</TABLE>\r\n");

	return 0;
}

static gint
draw_handlebox (FILE *output, int indent, xmlNode *handlebox)
{
	return draw_hbox (output, indent, handlebox);
}

static gint 
draw_toolbar (FILE *output, int indent, xmlNode *toolbar)
{
	return draw_hbox (output, indent, toolbar);
}

static gint
draw_form (FILE *output, int indent, xmlNode *form)
{
	xmlNode *child;
	char *title;

	title = get_node_content (get_node (form, "title"), "text");

	/* debugmsg ("Outputing form: %s", title); */

	output_html ("<HTML>\r\n");
	output_html ("<HEAD>\r\n");
	output_html ("<TITLE>%s</TITLE>\r\n", title);
	
	output_html ("<link rel=\"stylesheet\" type=\"text/css\" href=\"http://ignavus.net/netscape.css\">\r\n");
   output_html ("<link rel=\"stylesheet\" type=\"text/css\" href=\"http://ignavus.net/common.css\" media=\"all\">\r\n");
   output_html ("<link rel=\"stylesheet\" type=\"text/css\" href=\"http://ignavus.net/print.css\" media=\"print\">\r\n");

	output_html ("</HEAD>\r\n");
	output_html ("<BODY>\r\n");

	output_html ("<FORM>\r\n");

#warning "Not getting session details correctly"
	output_html ("<input type=\"hidden\" name=\"bond_session\" value=\"%d\">", 0);

	output_html ("<table> <!-- form table -->\r\n");

	for (child = form->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			{
			output_html ("<tr>\r\n");
			uihtml_draw_widget (output, indent + 1, child);
			output_html ("</tr>\r\n");
			}
		}

	output_html ("</table>\r\n");
	output_html ("</FORM>\r\n");

	output_html ("</BODY>\r\n");
	output_html ("</HTML>\r\n");

	return 0;
}

gint 
uihtml_draw_widget (FILE *output, int indent, xmlNode *widget)
{
	char *classname = NULL;
	xmlNode *class;

	for (class = widget->children; class != NULL; class = class->next)
		{
		if (strcmp (class->name, "class") == 0)
			classname = get_node_content (class, "text");
		}
	
/*	debugmsg ("Classname: %s", classname); */
	
	if (!classname)
		return -1;

	if (strcmp (classname, "GtkWindow") == 0)
		return draw_form (output, indent, widget);
	else if (strcmp (classname, "GtkLabel") == 0)
		return draw_label (output, indent, widget);
	else if (strcmp (classname, "GtkVBox") == 0)
		return draw_vbox (output, indent, widget);
	else if (strcmp (classname, "GtkHBox") == 0)
		return draw_hbox (output, indent, widget);
	else if (strcmp (classname, "GtkCList") == 0)
		return draw_clist (output, indent, widget);
	else if (strcmp (classname, "GtkScrolledWindow") == 0)
		return draw_empty (output, indent, widget);
	else if (strcmp (classname, "GtkTable") == 0)
		return draw_table (output, indent, widget);
	else if (strcmp (classname, "GtkEntry") == 0)
		return draw_entry (output, indent, widget);
	else if (strcmp (classname, "GtkCombo") == 0)
		return draw_combo (output, indent, widget);
	else if (strcmp (classname, "GtkButton") == 0)
		return draw_button (output, indent, widget);
	else if (strcmp (classname, "GtkText") == 0)
		return draw_text (output, indent, widget);
	else if (strcmp (classname, "GtkSpinButton") == 0)
		return draw_spin (output, indent, widget);
	else if (strcmp (classname, "GtkHandleBox") == 0)
		return draw_handlebox (output, indent, widget);
	else if (strcmp (classname, "GtkToolbar") == 0)
		return draw_toolbar (output, indent, widget);
	else if (strcmp (classname, "GtkCheckButton") == 0)
		return draw_checkbutton (output, indent, widget);
	else if (strcmp (classname, "GtkRadioButton") == 0)
		return draw_radiobutton (output, indent, widget);
	else if (strcmp (classname, "GtkHSeparator") == 0)
		return draw_separator (output, indent, widget);
	else if (strcmp (classname, "GtkAlignment") == 0)
		return draw_alignment (output, indent, widget);
	else if (strcmp (classname, "GtkPixmap") == 0)
		return draw_pixmap (output, indent, widget);
	else if (strcmp (classname, "Placeholder") == 0)
		return 0;
	
	errormsg ("Unknown class: %s", classname);
	return 0;
}

gint
uihtml_draw_form (xmlDoc *doc, int filedes, char *requiredname, int session)
{
	FILE *fp;
	char *formname;
	xmlNode *child, *name;
	int newfd;

	newfd = dup (filedes);

	if ((fp = fdopen (newfd, "w")) == NULL)
		{
		errormsg ("Unable to open file descriptor %d for writing", filedes);
		return -1;
		}

	for (child = doc->children->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			{
			name = get_node (child, "name");
			formname = get_node_content (name, "text");
			if (strcmp (formname, requiredname) == 0)
				{
				uihtml_draw_widget (fp, 0, child);
				break;
				}
			}
		}

		fclose (fp);

	return 0;
}

gint
uihtml_draw_all_forms (xmlDoc *doc)
{
	xmlNode *child;
	FILE *output;
	char *filename;
	char *formname;
	xmlNode *name;

	for (child = doc->children->children; child != NULL; child = child->next)
		{
		if (strcmp (child->name, "widget") == 0)
			{
			name = get_node (child, "name");
			formname = get_node_content (name, "text");
			asprintf (&filename, "html/%s.html", formname);

			if ((output = fopen (filename, "w")) != NULL)
				{
				uihtml_draw_widget (output, 0, child);	
				fclose (output);
				}
			else
				errormsg ("Unable to open %s: %s", filename, strerror (errno));

			free (filename);
			}
		else if (strcmp (child->name, "project") == 0)
			;
		else if (strcmp (child->name, "text") == 0)
			;
		else
			errormsg ("Unknown document node type: %s", child->name);
		}

	return 0;
}
